"""The task-based multi-process processing analysis engine."""

import collections
import os
import time

from plaso.containers import counts
from plaso.containers import events
from plaso.containers import reports
from plaso.containers import tasks
from plaso.engine import processing_status
from plaso.lib import definitions
from plaso.lib import errors
from plaso.multi_process import analysis_process
from plaso.multi_process import logger
from plaso.multi_process import merge_helpers
from plaso.multi_process import plaso_queue
from plaso.multi_process import task_engine
from plaso.multi_process import zeromq_queue


class AnalysisMultiProcessEngine(task_engine.TaskMultiProcessEngine):
  """Task-based multi-process analysis engine.

  This class contains functionality to:
  * monitor and manage analysis tasks;
  * merge results returned by analysis worker processes.
  """

  # pylint: disable=abstract-method

  _CONTAINER_TYPE_ANALYSIS_REPORT = reports.AnalysisReport.CONTAINER_TYPE
  _CONTAINER_TYPE_EVENT_TAG = events.EventTag.CONTAINER_TYPE

  _PROCESS_JOIN_TIMEOUT = 5.0

  _QUEUE_TIMEOUT = 10 * 60

  def __init__(self, worker_memory_limit=None, worker_timeout=None):
    """Initializes a task-based multi-process analysis engine.

    Args:
      worker_memory_limit (Optional[int]): maximum amount of memory a worker is
          allowed to consume, where None represents the default memory limit
          and 0 represents no limit.
      worker_timeout (Optional[float]): number of minutes before a worker
          process that is not providing status updates is considered inactive,
          where None or 0.0 represents the default timeout.
    """
    if worker_memory_limit is None:
      worker_memory_limit = definitions.DEFAULT_WORKER_MEMORY_LIMIT

    super().__init__()
    self._analysis_plugins = {}
    self._completed_analysis_processes = set()
    self._data_location = None
    self._event_filter_expression = None
    self._event_labels_counter = None
    self._event_queues = {}
    self._events_status = processing_status.EventsStatus()
    self._memory_profiler = None
    self._merge_task = None
    self._number_of_consumed_analysis_reports = 0
    self._number_of_consumed_event_data = 0
    self._number_of_consumed_event_tags = 0
    self._number_of_consumed_events = 0
    self._number_of_consumed_sources = 0
    self._number_of_produced_analysis_reports = 0
    self._number_of_produced_event_data = 0
    self._number_of_produced_event_tags = 0
    self._number_of_produced_events = 0
    self._number_of_produced_sources = 0
    self._processing_profiler = None
    self._serializers_profiler = None
    self._session = None
    self._status = definitions.STATUS_INDICATOR_IDLE
    self._status_update_callback = None
    self._user_accounts = None
    self._worker_memory_limit = worker_memory_limit
    self._worker_timeout = worker_timeout or definitions.DEFAULT_WORKER_TIMEOUT

  def _AnalyzeEvents(self, storage_writer, analysis_plugins, event_filter=None):
    """Analyzes events in a Plaso storage.

    Args:
      storage_writer (StorageWriter): storage writer.
      analysis_plugins (dict[str, AnalysisPlugin]): analysis plugins that
          should be run and their names.
      event_filter (Optional[EventObjectFilter]): event filter.

    Returns:
      collections.Counter: counter containing information about the events
          processed and filtered.

    Raises:
      RuntimeError: if a non-recoverable situation is encountered.
    """
    self._status = definitions.STATUS_INDICATOR_RUNNING
    self._number_of_consumed_analysis_reports = 0
    self._number_of_consumed_event_data = 0
    self._number_of_consumed_event_tags = 0
    self._number_of_consumed_events = 0
    self._number_of_consumed_sources = 0
    self._number_of_produced_analysis_reports = 0
    self._number_of_produced_event_data = 0
    self._number_of_produced_event_tags = 0
    self._number_of_produced_events = 0
    self._number_of_produced_sources = 0

    number_of_filtered_events = 0

    logger.debug('Processing events.')

    filter_limit = getattr(event_filter, 'limit', None)

    for event in storage_writer.GetSortedEvents():
      event_data_identifier = event.GetEventDataIdentifier()
      event_data = storage_writer.GetAttributeContainerByIdentifier(
          events.EventData.CONTAINER_TYPE, event_data_identifier)

      event_data_stream_identifier = event_data.GetEventDataStreamIdentifier()
      if event_data_stream_identifier:
        event_data_stream = storage_writer.GetAttributeContainerByIdentifier(
            events.EventDataStream.CONTAINER_TYPE, event_data_stream_identifier)
      else:
        event_data_stream = None

      event_identifier = event.GetIdentifier()
      event_tag = storage_writer.GetEventTagByEventIdentifer(event_identifier)

      if event_filter:
        filter_match = event_filter.Match(
            event, event_data, event_data_stream, event_tag)
      else:
        filter_match = None

      # pylint: disable=singleton-comparison
      if filter_match == False:
        number_of_filtered_events += 1
        continue

      for event_queue in self._event_queues.values():
        # TODO: Check for premature exit of analysis plugins.
        event_tripple = events.EventTripple()
        event_tripple.event = event
        event_tripple.event_data = event_data
        event_tripple.event_data_stream = event_data_stream

        event_queue.PushItem(event_tripple)

      self._number_of_consumed_events += 1

      if (event_filter and filter_limit and
          filter_limit == self._number_of_consumed_events):
        break

    logger.debug('Finished pushing events to analysis plugins.')
    # Signal that we have finished adding events.
    for event_queue in self._event_queues.values():
      event_queue.PushItem(plaso_queue.QueueAbort(), block=False)

    logger.debug('Processing analysis plugin results.')

    # TODO: use a task based approach.
    plugin_names = list(analysis_plugins.keys())
    while plugin_names:
      for plugin_name in list(plugin_names):
        if self._abort:
          break

        # TODO: temporary solution.
        task = tasks.Task()
        task.storage_format = definitions.STORAGE_FORMAT_SQLITE
        task.identifier = plugin_name

        merge_ready = self._CheckTaskReadyForMerge(
            definitions.STORAGE_FORMAT_SQLITE, task)
        if merge_ready:
          self._PrepareMergeTaskStorage(definitions.STORAGE_FORMAT_SQLITE, task)
          self._status = definitions.STATUS_INDICATOR_MERGING

          event_queue = self._event_queues[plugin_name]
          del self._event_queues[plugin_name]

          event_queue.Close()

          task_storage_reader = self._GetMergeTaskStorage(
              definitions.STORAGE_FORMAT_SQLITE, task)

          try:
            merge_helper = merge_helpers.AnalysisTaskMergeHelper(
                task_storage_reader, task.identifier)

            logger.debug(
                f'Starting merge of task: {merge_helper.task_identifier:s}')

            number_of_containers = self._MergeAttributeContainers(
                storage_writer, merge_helper)

            logger.debug((
                f'Merged {number_of_containers:d} containers of task: '
                f'{merge_helper.task_identifier:s}'))

          finally:
            task_storage_reader.Close()

          self._RemoveMergeTaskStorage(
              definitions.STORAGE_FORMAT_SQLITE, task)

          self._status = definitions.STATUS_INDICATOR_RUNNING

          # TODO: temporary solution.
          plugin_names.remove(plugin_name)

    events_counter = collections.Counter()
    events_counter['Events filtered'] = number_of_filtered_events
    events_counter['Events processed'] = self._number_of_consumed_events

    return events_counter

  def _CheckStatusAnalysisProcess(self, pid):
    """Checks the status of an analysis process.

    Args:
      pid (int): process ID (PID) of a registered analysis process.

    Raises:
      KeyError: if the process is not registered with the engine.
    """
    # TODO: Refactor this method, simplify and separate concerns (monitoring
    # vs management).
    self._RaiseIfNotRegistered(pid)

    if pid in self._completed_analysis_processes:
      status_indicator = definitions.STATUS_INDICATOR_COMPLETED
      process_status = {
          'processing_status': status_indicator}
      used_memory = 0

    else:
      process = self._processes_per_pid[pid]

      process_status = self._QueryProcessStatus(process)
      if process_status is None:
        process_is_alive = False
      else:
        process_is_alive = True

      process_information = self._process_information_per_pid[pid]
      used_memory = process_information.GetUsedMemory() or 0

      if self._worker_memory_limit and used_memory > self._worker_memory_limit:
        logger.warning((
            f'Process: {process.name:s} (PID: {pid:d}) killed because it '
            f'exceeded the memory limit: {self._worker_memory_limit:d}'))
        self._KillProcess(pid)

      if isinstance(process_status, dict):
        self._rpc_errors_per_pid[pid] = 0
        status_indicator = process_status.get('processing_status')

        if status_indicator == definitions.STATUS_INDICATOR_COMPLETED:
          self._completed_analysis_processes.add(pid)

      else:
        rpc_errors = self._rpc_errors_per_pid.get(pid, 0) + 1
        self._rpc_errors_per_pid[pid] = rpc_errors

        if rpc_errors > self._MAXIMUM_RPC_ERRORS:
          process_is_alive = False

        if process_is_alive:
          rpc_port = process.rpc_port.value
          logger.warning((
              f'Unable to retrieve process: {process.name:s} (PID: {pid:d}) '
              f'status via RPC socket: http://localhost:{rpc_port:d}'))

          processing_status_string = 'RPC error'
          status_indicator = definitions.STATUS_INDICATOR_RUNNING
        else:
          processing_status_string = 'killed'
          status_indicator = definitions.STATUS_INDICATOR_KILLED

        process_status = {
            'processing_status': processing_status_string}

    self._UpdateProcessingStatus(pid, process_status, used_memory)

    if status_indicator in definitions.ERROR_STATUS_INDICATORS:
      logger.error((
          f'Process {process.name:s} (PID: {pid:d}) is not functioning '
          f'correctly. Status code: {status_indicator!s}.'))

      self._TerminateProcessByPid(pid)

  def _MergeAttributeContainers(self, storage_writer, merge_helper):
    """Merges attribute containers from a task store into the storage writer.

    Args:
      storage_writer (StorageWriter): storage writer.
      merge_helper (AnalysisTaskMergeHelper): helper to merge attribute
          containers.

    Returns:
      int: number of containers merged.
    """
    number_of_containers = 0

    container = merge_helper.GetAttributeContainer()

    while container:
      number_of_containers += 1

      if container.CONTAINER_TYPE == self._CONTAINER_TYPE_EVENT_TAG:
        storage_writer.AddOrUpdateEventTag(container)
      else:
        storage_writer.AddAttributeContainer(container)

      if container.CONTAINER_TYPE == self._CONTAINER_TYPE_ANALYSIS_REPORT:
        self._number_of_produced_analysis_reports += 1

      elif container.CONTAINER_TYPE == self._CONTAINER_TYPE_EVENT_TAG:
        self._number_of_produced_event_tags += 1

        for label in container.labels:
          self._event_labels_counter[label] += 1
          self._event_labels_counter['total'] += 1

      container = merge_helper.GetAttributeContainer()

    return number_of_containers

  def _StartAnalysisProcesses(self, analysis_plugins):
    """Starts the analysis processes.

    Args:
      analysis_plugins (dict[str, AnalysisPlugin]): analysis plugins that
          should be run and their names.
    """
    logger.info('Starting analysis plugins.')

    for analysis_plugin in analysis_plugins.values():
      self._analysis_plugins[analysis_plugin.NAME] = analysis_plugin

      process = self._StartWorkerProcess(analysis_plugin.NAME)
      if not process:
        logger.error(
            f'Unable to create analysis process: {analysis_plugin.NAME:s}')

    logger.info('Analysis plugins running')

  def _StartWorkerProcess(self, process_name):
    """Creates, starts, monitors and registers a worker process.

    Args:
      process_name (str): process name.

    Returns:
      MultiProcessWorkerProcess: extraction worker process or None on error.
    """
    analysis_plugin = self._analysis_plugins.get(process_name)
    if not analysis_plugin:
      logger.error(f'Missing analysis plugin: {process_name:s}')
      return None

    queue_name = f'{process_name:s} output event queue'
    output_event_queue = zeromq_queue.ZeroMQPushBindQueue(
        name=queue_name, timeout_seconds=self._QUEUE_TIMEOUT)
    # Open the queue so it can bind to a random port, and we can get the
    # port number to use in the input queue.
    output_event_queue.Open()

    self._event_queues[process_name] = output_event_queue

    queue_name = f'{process_name:s} input event queue'
    input_event_queue = zeromq_queue.ZeroMQPullConnectQueue(
        name=queue_name, delay_open=True, port=output_event_queue.port,
        timeout_seconds=self._QUEUE_TIMEOUT)

    process = analysis_process.AnalysisProcess(
        input_event_queue, analysis_plugin, self._processing_configuration,
        self._user_accounts, data_location=self._data_location,
        event_filter_expression=self._event_filter_expression,
        name=process_name)

    process.start()

    logger.info(
        f'Started analysis plugin: {process_name:s} (PID: {process.pid:d}).')

    try:
      self._StartMonitoringProcess(process)
    except (KeyError, OSError) as exception:
      logger.error((
          f'Unable to monitor analysis plugin: {process_name:s} (PID: '
          f'{process.pid:d}) with error: {exception!s}'))

      process.terminate()
      return None

    self._RegisterProcess(process)
    return process

  def _StopAnalysisProcesses(self, abort=False):
    """Stops the analysis processes.

    Args:
      abort (bool): True to indicated the stop is issued on abort.
    """
    logger.debug('Stopping analysis processes.')
    self._StopMonitoringProcesses()

    if abort:
      # Signal all the processes to abort.
      self._AbortTerminate()

    # Wake the processes to make sure that they are not blocking
    # waiting for the queue new items.
    for event_queue in self._event_queues.values():
      event_queue.PushItem(plaso_queue.QueueAbort(), block=False)

    # Try waiting for the processes to exit normally.
    self._AbortJoin(timeout=self._PROCESS_JOIN_TIMEOUT)
    for event_queue in self._event_queues.values():
      event_queue.Close(abort=abort)

    if abort:
      # Kill any remaining processes.
      self._AbortKill()
    else:
      # Check if the processes are still alive and terminate them if necessary.
      self._AbortTerminate()
      self._AbortJoin(timeout=self._PROCESS_JOIN_TIMEOUT)

      for event_queue in self._event_queues.values():
        event_queue.Close(abort=True)

  def _UpdateForemanProcessStatus(self):
    """Update the foreman process status."""
    used_memory = self._process_information.GetUsedMemory() or 0

    display_name = getattr(self._merge_task, 'identifier', '')

    self._processing_status.UpdateForemanStatus(
        self._name, self._status, self._pid, used_memory, display_name,
        self._number_of_consumed_sources, self._number_of_produced_sources,
        self._number_of_consumed_event_data,
        self._number_of_produced_event_data,
        self._number_of_consumed_events, self._number_of_produced_events,
        self._number_of_consumed_event_tags,
        self._number_of_produced_event_tags,
        self._number_of_consumed_analysis_reports,
        self._number_of_produced_analysis_reports)

    self._processing_status.UpdateEventsStatus(self._events_status)

  def _UpdateProcessingStatus(self, pid, process_status, used_memory):
    """Updates the processing status.

    Args:
      pid (int): process identifier (PID) of the worker process.
      process_status (dict[str, object]): status values received from
          the worker process.
      used_memory (int): size of used memory in bytes.

    Raises:
      KeyError: if the process is not registered with the engine.
    """
    self._RaiseIfNotRegistered(pid)

    if not process_status:
      return

    process = self._processes_per_pid[pid]

    status_indicator = process_status.get('processing_status')

    self._RaiseIfNotMonitored(pid)

    display_name = process_status.get('display_name', '')

    number_of_consumed_event_data = process_status.get(
        'number_of_consumed_event_data', None)
    number_of_produced_event_data = process_status.get(
        'number_of_produced_event_data', None)

    number_of_consumed_event_tags = process_status.get(
        'number_of_consumed_event_tags', None)
    number_of_produced_event_tags = process_status.get(
        'number_of_produced_event_tags', None)

    number_of_consumed_events = process_status.get(
        'number_of_consumed_events', None)
    number_of_produced_events = process_status.get(
        'number_of_produced_events', None)

    number_of_consumed_reports = process_status.get(
        'number_of_consumed_reports', None)
    number_of_produced_reports = process_status.get(
        'number_of_produced_reports', None)

    number_of_consumed_sources = process_status.get(
        'number_of_consumed_sources', None)
    number_of_produced_sources = process_status.get(
        'number_of_produced_sources', None)

    if status_indicator != definitions.STATUS_INDICATOR_IDLE:
      last_activity_timestamp = process_status.get(
          'last_activity_timestamp', 0.0)

      if last_activity_timestamp:
        last_activity_timestamp += self._worker_timeout

        current_timestamp = time.time()
        if current_timestamp > last_activity_timestamp:
          logger.error((
              f'Process {process.name:s} (PID: {pid:d}) has not reported '
              f'activity within the timeout period.'))
          status_indicator = definitions.STATUS_INDICATOR_NOT_RESPONDING

    self._processing_status.UpdateWorkerStatus(
        process.name, status_indicator, pid, used_memory, display_name,
        number_of_consumed_sources, number_of_produced_sources,
        number_of_consumed_event_data, number_of_produced_event_data,
        number_of_consumed_events, number_of_produced_events,
        number_of_consumed_event_tags, number_of_produced_event_tags,
        number_of_consumed_reports, number_of_produced_reports)

  def _UpdateStatus(self):
    """Update the status."""
    # Make a local copy of the PIDs in case the dict is changed by
    # the main thread.
    for pid in list(self._process_information_per_pid.keys()):
      self._CheckStatusAnalysisProcess(pid)

    self._UpdateForemanProcessStatus()

    if self._status_update_callback:
      self._status_update_callback(self._processing_status)

  # pylint: disable=too-many-arguments
  def AnalyzeEvents(
      self, session, storage_writer, data_location, analysis_plugins,
      processing_configuration, event_filter=None,
      event_filter_expression=None, status_update_callback=None,
      storage_file_path=None):
    """Analyzes events in a Plaso storage.

    Args:
      session (Session): session in which the events are analyzed.
      storage_writer (StorageWriter): storage writer.
      data_location (str): path to the location that data files should
          be loaded from.
      analysis_plugins (dict[str, AnalysisPlugin]): analysis plugins that
          should be run and their names.
      processing_configuration (ProcessingConfiguration): processing
          configuration.
      event_filter (Optional[EventObjectFilter]): event filter.
      event_filter_expression (Optional[str]): event filter expression.
      status_update_callback (Optional[function]): callback function for status
          updates.
      storage_file_path (Optional[str]): path to the session storage file.

    Returns:
      ProcessingStatus: processing status.

    Raises:
      KeyboardInterrupt: if a keyboard interrupt was raised.
      ValueError: if analysis plugins are missing.
    """
    if not analysis_plugins:
      raise ValueError('Missing analysis plugins')

    abort_kill = False
    keyboard_interrupt = False
    queue_full = False

    self._analysis_plugins = {}
    self._data_location = data_location
    self._event_filter_expression = event_filter_expression
    self._events_status = processing_status.EventsStatus()
    self._processing_configuration = processing_configuration
    self._session = session
    self._status_update_callback = status_update_callback
    self._storage_file_path = storage_file_path

    self._user_accounts = list(
        storage_writer.GetAttributeContainers('user_account'))

    stored_event_labels_counter = {}
    if storage_writer.HasAttributeContainers('event_label_count'):
      stored_event_labels_counter = {
          event_label_count.label: event_label_count
          for event_label_count in storage_writer.GetAttributeContainers(
              'event_label_count')}

    self._event_labels_counter = collections.Counter()

    if storage_writer.HasAttributeContainers('parser_count'):
      parsers_counter = {
          parser_count.name: parser_count.number_of_events
          for parser_count in storage_writer.GetAttributeContainers(
              'parser_count')}

      total_number_of_events = parsers_counter['total']

    else:
      total_number_of_events = 0
      for stored_session in storage_writer.GetSessions():
        total_number_of_events += stored_session.parsers_counter['total']

    self._events_status.total_number_of_events = total_number_of_events

    # Set up the storage writer before the analysis processes.
    self._StartTaskStorage(definitions.STORAGE_FORMAT_SQLITE)

    self._StartAnalysisProcesses(analysis_plugins)

    self._StartProfiling(self._processing_configuration.profiling)

    # Start the status update thread after open of the storage writer
    # so we don't have to clean up the thread if the open fails.
    self._StartStatusUpdateThread()

    try:
      self._AnalyzeEvents(
          storage_writer, analysis_plugins, event_filter=event_filter)

      for key, value in self._event_labels_counter.items():
        event_label_count = stored_event_labels_counter.get(key)
        if event_label_count:
          event_label_count.number_of_events += value
          storage_writer.UpdateAttributeContainer(event_label_count)
        else:
          event_label_count = counts.EventLabelCount(
              label=key, number_of_events=value)
          storage_writer.AddAttributeContainer(event_label_count)

      self._status = definitions.STATUS_INDICATOR_FINALIZING

    except errors.QueueFull:
      queue_full = True
      self._abort = True

    except KeyboardInterrupt:
      keyboard_interrupt = True
      self._abort = True

    finally:
      self._processing_status.aborted = self._abort
      session.aborted = self._abort

      # Stop the status update thread after close of the storage writer
      # so we include the storage sync to disk in the status updates.
      self._StopStatusUpdateThread()

      self._StopProfiling()

    # Update the status view one last time before the analysis processses are
    # stopped.
    self._UpdateStatus()

    if queue_full:
      # TODO: handle abort on queue full more elegant.
      abort_kill = True
    else:
      try:
        self._StopAnalysisProcesses(abort=self._abort)

      except KeyboardInterrupt:
        keyboard_interrupt = True
        abort_kill = True

    if abort_kill:
      self._AbortKill()

      # The abort can leave the main process unresponsive
      # due to incorrectly finalized IPC.
      self._KillProcess(os.getpid())

    try:
      self._StopTaskStorage(
          definitions.STORAGE_FORMAT_SQLITE, session.identifier,
          abort=self._abort)
    except OSError as exception:
      logger.error(f'Unable to stop task storage with error: {exception!s}')

    if self._abort:
      logger.debug('Analysis aborted.')
    else:
      logger.debug('Analysis completed.')

    # Update the status view one last time.
    self._UpdateStatus()

    # Reset values.
    self._analysis_plugins = {}
    self._data_location = None
    self._event_filter_expression = None
    self._processing_configuration = None
    self._session = None
    self._status_update_callback = None
    self._storage_file_path = None
    self._user_accounts = None

    if keyboard_interrupt:
      raise KeyboardInterrupt

    return self._processing_status



"""SQLite parser plugin for Google Chrome autofill database files.

The Google Chrome autofill database (Web Data) file is typically stored in:
  Web Data
"""

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class ChromeAutofillEventData(events.EventData):
  """Chrome Autofill event data.

  Attributes:
    creation_time (dfdatetime.DateTimeValues): creation date and time of
        the autofill entry.
    field_name (str): name of form field.
    last_used_time (dfdatetime.DateTimeValues): last date and time
        the autofill entry was last used.
    query (str): SQL query that was used to obtain the event data.
    usage_count (int): count of times value has been used in field_name.
    value (str): value populated in form field.
  """

  DATA_TYPE = 'chrome:autofill:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.creation_time = None
    self.field_name = None
    self.last_used_time = None
    self.query = None
    self.usage_count = None
    self.value = None


class ChromeAutofillPlugin(interface.SQLitePlugin):
  """SQLite parser plugin for Google Chrome autofill database files."""

  NAME = 'chrome_autofill'
  DATA_FORMAT = 'Google Chrome autofill SQLite database (Web Data) file'

  REQUIRED_STRUCTURE = {
      'autofill': frozenset([
          'count', 'date_created', 'date_last_used', 'name', 'value'])}

  QUERIES = [(
      ('SELECT count, date_created, date_last_used, name, value '
       'FROM autofill'), '_ParseAutofillRow')]

  SCHEMAS = [{
      'autofill': (
          'CREATE TABLE autofill (name VARCHAR, value VARCHAR, value_lower '
          'VARCHAR, date_created INTEGER DEFAULT 0, date_last_used INTEGER '
          'DEFAULT 0, count INTEGER DEFAULT 1, PRIMARY KEY (name, value))'),
      'autofill_model_type_state': (
          'CREATE TABLE autofill_model_type_state (id INTEGER PRIMARY KEY, '
          'value BLOB)'),
      'autofill_profile_emails': (
          'CREATE TABLE autofill_profile_emails ( guid VARCHAR, email '
          'VARCHAR)'),
      'autofill_profile_names': (
          'CREATE TABLE autofill_profile_names ( guid VARCHAR, first_name '
          'VARCHAR, middle_name VARCHAR, last_name VARCHAR, full_name '
          'VARCHAR)'),
      'autofill_profile_phones': (
          'CREATE TABLE autofill_profile_phones ( guid VARCHAR, number '
          'VARCHAR)'),
      'autofill_profiles': (
          'CREATE TABLE autofill_profiles ( guid VARCHAR PRIMARY KEY, '
          'company_name VARCHAR, street_address VARCHAR, dependent_locality '
          'VARCHAR, city VARCHAR, state VARCHAR, zipcode VARCHAR, '
          'sorting_code VARCHAR, country_code VARCHAR, date_modified INTEGER '
          'NOT NULL DEFAULT 0, origin VARCHAR DEFAULT \'\', language_code '
          'VARCHAR, use_count INTEGER NOT NULL DEFAULT 0, use_date INTEGER '
          'NOT NULL DEFAULT 0, validity_bitfield UNSIGNED NOT NULL DEFAULT 0)'),
      'autofill_profiles_trash': (
          'CREATE TABLE autofill_profiles_trash ( guid VARCHAR)'),
      'autofill_sync_metadata': (
          'CREATE TABLE autofill_sync_metadata (storage_key VARCHAR PRIMARY '
          'KEY NOT NULL,value BLOB)'),
      'credit_cards': (
          'CREATE TABLE credit_cards ( guid VARCHAR PRIMARY KEY, name_on_card '
          'VARCHAR, expiration_month INTEGER, expiration_year INTEGER, '
          'card_number_encrypted BLOB, date_modified INTEGER NOT NULL DEFAULT '
          '0, origin VARCHAR DEFAULT \'\', use_count INTEGER NOT NULL DEFAULT '
          '0, use_date INTEGER NOT NULL DEFAULT 0, billing_address_id '
          'VARCHAR)'),
      'keywords': (
          'CREATE TABLE keywords (id INTEGER PRIMARY KEY,short_name VARCHAR '
          'NOT NULL,keyword VARCHAR NOT NULL,favicon_url VARCHAR NOT NULL,url '
          'VARCHAR NOT NULL,safe_for_autoreplace INTEGER,originating_url '
          'VARCHAR,date_created INTEGER DEFAULT 0,usage_count INTEGER DEFAULT '
          '0,input_encodings VARCHAR,suggest_url VARCHAR,prepopulate_id '
          'INTEGER DEFAULT 0,created_by_policy INTEGER DEFAULT '
          '0,last_modified INTEGER DEFAULT 0,sync_guid VARCHAR,alternate_urls '
          'VARCHAR,image_url VARCHAR,search_url_post_params '
          'VARCHAR,suggest_url_post_params VARCHAR,image_url_post_params '
          'VARCHAR,new_tab_url VARCHAR,last_visited INTEGER DEFAULT 0)'),
      'masked_credit_cards': (
          'CREATE TABLE masked_credit_cards (id VARCHAR,status '
          'VARCHAR,name_on_card VARCHAR,network VARCHAR,last_four '
          'VARCHAR,exp_month INTEGER DEFAULT 0,exp_year INTEGER DEFAULT 0, '
          'bank_name VARCHAR, type INTEGER DEFAULT 0)'),
      'meta': (
          'CREATE TABLE meta(key LONGVARCHAR NOT NULL UNIQUE PRIMARY KEY, '
          'value LONGVARCHAR)'),
      'payment_method_manifest': (
          'CREATE TABLE payment_method_manifest ( expire_date INTEGER NOT '
          'NULL DEFAULT 0, method_name VARCHAR, web_app_id VARCHAR)'),
      'server_address_metadata': (
          'CREATE TABLE server_address_metadata (id VARCHAR NOT '
          'NULL,use_count INTEGER NOT NULL DEFAULT 0, use_date INTEGER NOT '
          'NULL DEFAULT 0, has_converted BOOL NOT NULL DEFAULT FALSE)'),
      'server_addresses': (
          'CREATE TABLE server_addresses (id VARCHAR,company_name '
          'VARCHAR,street_address VARCHAR,address_1 VARCHAR,address_2 '
          'VARCHAR,address_3 VARCHAR,address_4 VARCHAR,postal_code '
          'VARCHAR,sorting_code VARCHAR,country_code VARCHAR,language_code '
          'VARCHAR, recipient_name VARCHAR, phone_number VARCHAR)'),
      'server_card_metadata': (
          'CREATE TABLE server_card_metadata (id VARCHAR NOT NULL,use_count '
          'INTEGER NOT NULL DEFAULT 0, use_date INTEGER NOT NULL DEFAULT 0, '
          'billing_address_id VARCHAR)'),
      'unmasked_credit_cards': (
          'CREATE TABLE unmasked_credit_cards (id '
          'VARCHAR,card_number_encrypted VARCHAR, use_count INTEGER NOT NULL '
          'DEFAULT 0, use_date INTEGER NOT NULL DEFAULT 0, unmask_date '
          'INTEGER NOT NULL DEFAULT 0)'),
      'web_app_manifest_section': (
          'CREATE TABLE web_app_manifest_section ( expire_date INTEGER NOT '
          'NULL DEFAULT 0, id VARCHAR, min_version INTEGER NOT NULL DEFAULT '
          '0, fingerprints BLOB)')}]

  def _GetDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.PosixTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_posix_time.PosixTime(timestamp=timestamp)

  def _ParseAutofillRow(self, parser_mediator, query, row, **unused_kwargs):
    """Parses an autofill entry row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    event_data = ChromeAutofillEventData()
    event_data.creation_time = self._GetDateTimeRowValue(
        query_hash, row, 'date_created')
    event_data.field_name = self._GetRowValue(query_hash, row, 'name')
    event_data.last_used_time = self._GetDateTimeRowValue(
        query_hash, row, 'date_last_used')
    event_data.query = query
    event_data.usage_count = self._GetRowValue(query_hash, row, 'count')
    event_data.value = self._GetRowValue(query_hash, row, 'value')

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(ChromeAutofillPlugin)



#!/usr/bin/env python3
"""The log2timeline command line tool."""

import logging
import multiprocessing
import os
import sys

from plaso import dependencies
from plaso.cli import log2timeline_tool
from plaso.lib import errors


def Main():
  """Entry point of console script to extract events.

  Returns:
    int: exit code that is provided to sys.exit().
  """
  tool = log2timeline_tool.Log2TimelineTool()

  if not tool.ParseArguments(sys.argv[1:]):
    return 1

  if tool.show_troubleshooting:
    path = os.path.abspath(__file__)

    print(f'Using Python version {sys.version!s}')
    print()
    print(f'Path: {path!s}')
    print()
    print(tool.GetVersionInformation())
    print()
    dependencies.CheckDependencies(verbose_output=True)

    print('Also see: https://plaso.readthedocs.io/en/latest/sources/user/'
          'Troubleshooting.html')
    return 0

  try:
    tool.CheckOutDated()
  except KeyboardInterrupt:
    return 1

  if tool.show_info:
    tool.ShowInfo()
    return 0

  have_list_option = False
  if tool.list_archive_types:
    tool.ListArchiveTypes()
    have_list_option = True

  if tool.list_hashers:
    tool.ListHashers()
    have_list_option = True

  if tool.list_language_tags:
    tool.ListLanguageTags()
    have_list_option = True

  if tool.list_parsers_and_plugins:
    tool.ListParsersAndPlugins()
    have_list_option = True

  if tool.list_profilers:
    tool.ListProfilers()
    have_list_option = True

  if tool.list_time_zones:
    tool.ListTimeZones()
    have_list_option = True

  if have_list_option:
    return 0

  if tool.dependencies_check and not dependencies.CheckDependencies(
      verbose_output=False):
    return 1

  try:
    tool.ExtractEventsFromSources()

  # Writing to stdout and stderr will raise BrokenPipeError if it
  # receives a SIGPIPE.
  except BrokenPipeError:
    pass

  except (KeyboardInterrupt, errors.UserAbort):
    logging.warning('Aborted by user.')
    return 1

  except (OSError, errors.BadConfigOption,
          errors.SourceScannerError) as exception:
    # Display message on stdout as well as the log file.
    print(exception)
    logging.error(exception)
    return 1

  return 0


if __name__ == '__main__':
  multiprocessing.freeze_support()

  sys.exit(Main())



"""Plist parser plugin for Spotlight volume configuration plist files."""

from plaso.containers import events
from plaso.parsers import plist
from plaso.parsers.plist_plugins import interface


class SpotlightVolumeConfigurationEventData(events.EventData):
  """Spotlight volume configuration event data.

  Attributes:
    creation_time (dfdatetime.DateTimeValues): volume creation date and time.
    partial_path (str): part of the path.
    volume_identifier (str): identifier of the volume.
  """

  DATA_TYPE = 'spotlight_volume_configuration:store'

  def __init__(self):
    """Initializes event data."""
    super().__init__(
        data_type=self.DATA_TYPE)
    self.creation_time = None
    self.partial_path = None
    self.volume_identifier = None


class SpotlightVolumeConfigurationPlistPlugin(interface.PlistPlugin):
  """Plist parser plugin for Spotlight volume configuration plist files."""

  NAME = 'spotlight_volume'
  DATA_FORMAT = 'Spotlight volume configuration plist file'

  PLIST_PATH_FILTERS = frozenset([
      interface.PlistPathFilter('VolumeConfiguration.plist')])

  PLIST_KEYS = frozenset(['Stores'])

  # pylint: disable=arguments-differ
  def _ParsePlist(self, parser_mediator, match=None, **unused_kwargs):
    """Extracts relevant Volume Configuration Spotlight entries.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      match (Optional[dict[str: object]]): keys extracted from PLIST_KEYS.
    """
    stores = match.get('Stores', {})
    for volume_identifier, plist_key in stores.items():
      event_data = SpotlightVolumeConfigurationEventData()
      event_data.creation_time = self._GetDateTimeValueFromPlistKey(
          plist_key, 'CreationDate')
      event_data.partial_path = plist_key.get('PartialPath')
      event_data.volume_identifier = volume_identifier

      parser_mediator.ProduceEventData(event_data)


plist.PlistParser.RegisterPlugin(SpotlightVolumeConfigurationPlistPlugin)



"""Value types that can be used in an event filter."""

from dfdatetime import posix_time as dfdatetime_posix_time


class DateTimeValueType(dfdatetime_posix_time.PosixTimeInMicroseconds):
  """Value type to represent a date and time value."""

  def __init__(self, value):
    """Initializes a date and time value type.

    Args:
      value (int:str): a POSIX timestamp in microseconds or an ISO 8601 date
          and time string.

    Raises:
      ValueError: if the value cannot be copied to a date and time object.
    """
    if isinstance(value, int):
      timestamp = value
    else:
      try:
        timestamp = int(value, 10)
      except (TypeError, ValueError):
        timestamp = None

    super().__init__(timestamp=timestamp)

    if timestamp is None:
      try:
        # Adjust the ISO 8601 date and time string so is resembles a Python
        # date and time string.
        if value and len(value) > 10 and value[10] == 'T':
          value = ' '.join(value.split('T'))

        self.CopyFromDateTimeString(value)
      except (TypeError, ValueError):
        raise ValueError(f'Unsupported date time string value: {value!s}')



"""Trie structure for storing artifact paths."""

import fnmatch
import glob
import os

from plaso.engine import logger
from plaso.engine import path_helper


class TrieNode:
  """Represents a node in the Trie.

  Attributes:
    artifacts_names (list[str]): Names of artifacts associated with this node.
    children (dict[str, TrieNode]): Child nodes, keyed by path segment.
    is_root (bool): True if this is the root node.
    path_separator (str): Path separator used in the Trie, default is '/'.
  """

  def __init__(self, path_separator='/', is_root=False):
    """Initializes a trie node object.

    Args:
      path_separator (Optional[str]): the path separator used in paths stored
          in the Trie, default is '/'.
      is_root (Optional[bool]): True if this node is the root node.
    """
    super().__init__()
    self.artifacts_names = []
    self.children = {}
    self.is_root = is_root
    self.path_separator = path_separator


class ArtifactsTrie:
  """Trie structure for storing artifact paths.

  Attributes:
    artifacts_paths (dict[str, list[str]]): Artifact paths for glob expansion,
        keyed by artifact name.
    root (TrieNode): Root node of the Trie.
  """

  def __init__(self):
    """Initializes an artifact trie object."""
    super().__init__()
    self.artifacts_paths = {}
    self.root = TrieNode(is_root=True)

  def AddPath(self, artifact_name, path, path_separator):
    """Adds a path from an artifact definition to the Trie.

    Args:
      artifact_name (str): name of the artifact.
      path (str): path from the artifact definition.
      path_separator (str): path separator.
    """
    logger.debug(f'Adding path: "{path:s}" to artifact: "{artifact_name:s}"')
    path_list = self.artifacts_paths.setdefault(artifact_name, [])
    path_list.append(path)

    current_node = self.root

    # Add a path separator node if this is a new separator.
    if path_separator not in current_node.children:
      current_node.children[path_separator] = TrieNode(
          path_separator=path_separator)
    current_node = current_node.children[path_separator]

    # Handle the case when the input path is equal to the path separator.
    if path == path_separator:
      current_node.artifacts_names.append(artifact_name)
      return

    path_segments = path.strip(path_separator).split(path_separator)
    for path_segment in path_segments:
      # Store the path_separator for each node.
      if not hasattr(current_node, 'path_separator'):
        current_node.path_separator = path_separator

      if path_segment not in current_node.children:
        current_node.children[path_segment] = TrieNode(path_separator)
      current_node = current_node.children[path_segment]
    current_node.artifacts_names.append(artifact_name)

  def GetMatchingArtifacts(self, path, path_separator):
    """Retrieves the artifact names that match the given path.

    Args:
        path (str): path to match against the Trie.
        path_separator (str): path separator.

    Returns:
      list[str]: artifact names that match the path.
    """
    # Start at the root's child that matches the path_separator.
    if path_separator not in self.root.children:
      return []

    sub_root_node = self.root.children[path_separator]

    # Handle the case when the input path is equal to the path_separator.
    if path == path_separator:
      matching_artifacts = set()
      if sub_root_node.artifacts_names:
        matching_artifacts.update(sub_root_node.artifacts_names)
      return list(matching_artifacts)
    path_segments = path.strip(path_separator).split(path_separator)
    matching_artifacts = set()
    # Update self.artifacts_paths before starting the search.
    self.artifacts_paths = self._GetArtifactsPaths(sub_root_node)

    self._SearchTrie(
      sub_root_node, '', path_segments, path_separator, matching_artifacts)
    return list(matching_artifacts)

  def _SearchTrie(
      self,node, current_path, segments, path_separator, matching_artifacts):
    """Searches the trie for paths matching the given path segments.

    Args:
      node (TrieNode): current trie node being traversed.
      current_path (str): path represented by the current node.
      segments (list[str]): remaining path segments to match.
      path_separator (str): path separator.
      matching_artifacts (set[str]): Set to store matching artifact names.
    """
    if node.artifacts_names:
      for artifact_name in node.artifacts_names:
        for artifact_path in self.artifacts_paths.get(artifact_name, []):
          # Note that the sanitized path is used here to ensure consistency
          # with the exported path.
          # TODO: consider moving path sanitation out of engine.
          if self._ComparePathIfSanitized(
              current_path, path_separator, artifact_path,
              node.path_separator):
            matching_artifacts.add(artifact_name)
          elif glob.has_magic(artifact_path):
            if self._MatchesGlobPattern(
                artifact_path, current_path, node.path_separator):
              matching_artifacts.add(artifact_name)

    if not segments:
      return

    segment = segments[0]
    remaining_segments = segments[1:]

    # Handle glob characters in the current segment.
    for child_segment, child_node in node.children.items():
      # Compare the sanitized version of the path segment stored in
      # the tree to the path segment from to the tool output as it
      # sanitizes path segments before writing data to disk.
      sanitized_child_segment = path_helper.PathHelper.SanitizePathSegments(
          [child_segment]).pop()

      # If the child is an exact match, continue traversal.
      if segment in (child_segment, sanitized_child_segment):
        custom_path = self._CustomPathJoin(
            path_separator, current_path, child_segment)
        self._SearchTrie(
            child_node, custom_path, remaining_segments, path_separator,
            matching_artifacts)
      # If the child is a glob, see if it matches.
      elif glob.has_magic(child_segment):
        if self._MatchesGlobPattern(
                child_segment, segment, child_node.path_separator):
          custom_path = self._CustomPathJoin(
              path_separator, current_path, segment)
          self._SearchTrie(
              child_node, custom_path, remaining_segments, path_separator,
              matching_artifacts)
          self._SearchTrie(
              node, custom_path, remaining_segments, path_separator,
              matching_artifacts)

  def _ComparePathIfSanitized(
      self, current_path, path_separator, artifact_path,
      artifact_path_seperator):
    """Compares a current path with an artifact path, handling sanitization.

    This method checks if the current_path matches the artifact_path,
    considering that the artifact_path might have been sanitized.

    Args:
      current_path (str): The current path being checked.
      path_separator (str): Path separator for the current path.
      artifact_path (str): The artifact path to compare against.
      artifact_path_seperator (str): Path separator for the artifact path.

    Returns:
      bool: True if the current path matches the artifact path or its
          sanitized version, False otherwise.
    """
    artifact_path_segments = self._GetNonEmptyPathSegments(
        artifact_path, artifact_path_seperator)
    sanitized_path_segments = path_helper.PathHelper.SanitizePathSegments(
        artifact_path_segments)
    return self._GetNonEmptyPathSegments(current_path, path_separator) in [
        artifact_path_segments, sanitized_path_segments]

  def _GetNonEmptyPathSegments(self, path, separator):
    """Splits a path into segments and remove non-empty segments.

    Args:
      path (str): The path string to be split.
      separator (str): The path separator.

    Returns:
      list[str]: A list of non-empty path segments.
    """
    return [s for s in path.split(separator) if s]

  def _GetArtifactsPaths(self, node):
    """Retrieves a mapping of artifact names to their paths.

    Args:
      node (TrieNode): current trie node being traversed.

    Returns:
      dict: dictionary mapping artifact names to their paths.
    """
    artifacts_paths = {}
    self._collect_paths(node, '', artifacts_paths)
    return artifacts_paths

  def _collect_paths(self, node, current_path, artifacts):
    """Collects paths from the trie.

    Args:
      node (TrieNode): current node.
      current_path (str): path leading to this node.
      artifacts (dict): dictionary to store artifact paths.
    """
    if node.artifacts_names:
      for artifact_name in node.artifacts_names:
        path_list = artifacts.setdefault(artifact_name, [])
        path_list.append(current_path)

    for segment, child_node in node.children.items():
      # Ensure the path_separator attribute exists.
      if not hasattr(child_node, 'path_separator'):
        child_node.path_separator = node.path_separator

      # Construct the next path segment.
      if current_path == child_node.path_separator:
        # Means it is the root folder, i.e. `/`
        next_path = current_path + segment
      else:
        next_path = current_path + child_node.path_separator + segment

      self._collect_paths(child_node, next_path, artifacts)

  def _CustomPathJoin(self, separator, current_path, new_segment):
    """Joins path components using a custom separator, replacing os.sep.

    Args:
      separator (str): The custom separator to use.
      current_path (str): The current path.
      new_segment (str): The new segment to add to it.

    Returns:
      str: The joined path with the custom separator.
    """
    current_path = current_path.replace(separator, os.sep)
    joined_path = os.path.join(current_path, new_segment)
    return joined_path.replace(os.sep, separator)

  def _MatchesGlobPattern(self, glob_pattern, path, path_separator):
    """Checks if a path matches a given glob pattern.

    Args:
      glob_pattern (str): The glob pattern to match against.
      path (str): The path to check.
      path_separator (str): The path separator used in the glob pattern.

    Returns:
      True if the path matches the glob pattern, False otherwise.
    """
    # Normalize paths using the appropriate separators
    glob_pattern = glob_pattern.strip(path_separator).split(path_separator)
    path = path.strip(path_separator).split(path_separator)

    glob_index = 0
    path_index = 0
    while glob_index < len(glob_pattern) and path_index < len(path):
      if glob_pattern[glob_index] == '**':
        # If ** is the last part, it matches everything remaining
        if glob_index == len(glob_pattern) - 1:
          return True
        # Move to the next part after **
        glob_index += 1
        # Keep advancing in the path until the next part matches
        while path_index < len(path) and not fnmatch.fnmatch(
          path[path_index], glob_pattern[glob_index]):
          path_index += 1
      elif not fnmatch.fnmatch(path[path_index], glob_pattern[glob_index]):
        # Mismatch
        return False
      else:
        glob_index += 1
        path_index += 1

    return glob_index == len(glob_pattern) and path_index == len(path)



"""Helper to create filters based on forensic artifact definitions."""

import os

from artifacts import definitions as artifact_types

from dfvfs.helpers import file_system_searcher as dfvfs_file_system_searcher

from dfwinreg import registry_searcher as dfwinreg_registry_searcher

from plaso.engine import artifacts_trie
from plaso.engine import logger
from plaso.engine import path_helper


class ArtifactDefinitionsFiltersHelper:
  """Helper to create collection filters based on artifact definitions.

  Builds collection filters from forensic artifact definitions.

  For more information about Forensic Artifacts see:
  https://github.com/ForensicArtifacts/artifacts/blob/main/docs/Artifacts%20definition%20format%20and%20style%20guide.asciidoc

  Attributes:
    artifacts_trie (ArtifactsTrie): Trie structure for storing artifact
        definition paths.
    file_system_artifact_names (set[str]): names of artifacts definitions that
        generated file system find specifications.
    file_system_find_specs (list[dfvfs.FindSpec]): file system find
        specifications of paths to include in the collection.
    registry_artifact_names (set[str]): names of artifacts definitions that
        generated Windows Registry find specifications.
    registry_find_specs (list[dfwinreg.FindSpec]): Windows Registry find
        specifications.
    registry_find_specs_artifact_names (list[str]): Windows Registry artifact
        names corresponding to the find specifications.
  """

  _COMPATIBLE_REGISTRY_KEY_PATH_PREFIXES = frozenset([
      'HKEY_CURRENT_USER',
      'HKEY_LOCAL_MACHINE\\SYSTEM',
      'HKEY_LOCAL_MACHINE\\SOFTWARE',
      'HKEY_LOCAL_MACHINE\\SAM',
      'HKEY_LOCAL_MACHINE\\SECURITY',
      'HKEY_USERS'])

  def __init__(self, artifacts_registry):
    """Initializes an artifact definitions filters helper.

    Args:
      artifacts_registry (artifacts.ArtifactDefinitionsRegistry): artifact
          definitions registry.
    """
    super().__init__()
    self._artifacts_registry = artifacts_registry

    self.artifacts_trie = artifacts_trie.ArtifactsTrie()
    self.file_system_artifact_names = set()
    self.file_system_find_specs = []
    self.registry_artifact_names = set()
    self.registry_find_specs = []
    self.registry_find_specs_artifact_names = []

  def _BuildFindSpecsFromArtifact(
      self, definition, environment_variables, user_accounts,
      enable_artifacts_map=False,
      original_registry_artifact_filter_names=None):
    """Builds find specifications from an artifact definition.

    Args:
      definition (artifacts.ArtifactDefinition): artifact definition.
      environment_variables (list[EnvironmentVariableArtifact]): environment
          variables.
      user_accounts (list[UserAccountArtifact]): user accounts.
      enable_artifacts_map (Optional[bool]): True if the artifacts path map
          should be generated. Defaults to False.
      original_registry_artifact_filter_names (Optional[set[str]]): Set of
          original registry filter names, used in case Windows Registry hive
          files are being requested as a result of a previous filter.

    Returns:
      list[dfvfs.FindSpec|dfwinreg.FindSpec]: dfVFS or dfWinReg find
          specifications.
    """
    find_specs = []
    for source in definition.sources:
      if source.type_indicator == artifact_types.TYPE_INDICATOR_FILE:
        for path_entry in set(source.paths):
          specifications = self._BuildFindSpecsFromFileSourcePath(
              definition.name, path_entry, source.separator,
              environment_variables, user_accounts,
              enable_artifacts_map=enable_artifacts_map,
              original_registry_artifact_filter_names=(
                  original_registry_artifact_filter_names))
          find_specs.extend(specifications)
          self.file_system_artifact_names.add(definition.name)

      elif (source.type_indicator ==
            artifact_types.TYPE_INDICATOR_WINDOWS_REGISTRY_KEY):
        for key_path in set(source.keys):
          if ArtifactDefinitionsFiltersHelper.CheckKeyCompatibility(key_path):
            specifications = self._BuildFindSpecsFromRegistrySourceKey(key_path)
            find_specs.extend(specifications)
            self.registry_artifact_names.add(definition.name)

      elif (source.type_indicator ==
            artifact_types.TYPE_INDICATOR_WINDOWS_REGISTRY_VALUE):
        # TODO: Handle Registry Values Once Supported in dfwinreg.
        # https://github.com/log2timeline/dfwinreg/issues/98

        # Use set-comprehension to create a set of the source key paths.
        key_paths = {key_value['key'] for key_value in source.key_value_pairs}
        key_paths_string = ', '.join(key_paths)

        logger.warning((
            f'Windows Registry values are not supported, extracting keys: '
            f'"{key_paths_string!s}"'))

        for key_path in key_paths:
          if ArtifactDefinitionsFiltersHelper.CheckKeyCompatibility(key_path):
            specifications = self._BuildFindSpecsFromRegistrySourceKey(key_path)
            find_specs.extend(specifications)
            self.registry_artifact_names.add(definition.name)

      elif (source.type_indicator ==
            artifact_types.TYPE_INDICATOR_ARTIFACT_GROUP):
        for name in source.names:
          specifications = self._BuildFindSpecsFromGroupName(
              name, environment_variables, user_accounts,
              enable_artifacts_map=enable_artifacts_map,
              original_registry_artifact_filter_names=(
                  original_registry_artifact_filter_names))
          find_specs.extend(specifications)

      else:
        logger.warning((
            f'Unsupported artifact definition source type: '
            f'"{source.type_indicator:s}"'))

    return find_specs

  def _BuildFindSpecsFromGroupName(
      self, group_name, environment_variables, user_accounts,
      enable_artifacts_map=False,
      original_registry_artifact_filter_names=None):
    """Builds find specifications from a artifact group name.

    Args:
      group_name (str): artifact group name.
      environment_variables (list[EnvironmentVariableArtifact]): environment
          variables.
      user_accounts (list[UserAccountArtifact]): user accounts.
      enable_artifacts_map (Optional[bool]): True if the artifacts path map
          should be generated. Defaults to False.
      original_registry_artifact_filter_names (Optional[set[str]]): Set of
          original registry filter names, used in case registry hive files
          are being requested as a result of a previous filter.

    Returns:
      list[dfwinreg.FindSpec|dfvfs.FindSpec]: find specifications or None if no
          artifact with the given name can be retrieved.
    """
    definition = self._artifacts_registry.GetDefinitionByName(group_name)
    if not definition:
      definition = self._artifacts_registry.GetDefinitionByAlias(group_name)
    if not definition:
      return None

    return self._BuildFindSpecsFromArtifact(
        definition, environment_variables, user_accounts,
        enable_artifacts_map=enable_artifacts_map,
        original_registry_artifact_filter_names=(
            original_registry_artifact_filter_names))

  def _BuildFindSpecsFromRegistrySourceKey(self, key_path):
    """Build find specifications from a Windows Registry source type.

    Args:
      key_path (str): Windows Registry key path defined by the source.

    Returns:
      list[dfwinreg.FindSpec]: find specifications for the Windows Registry
          source type.
    """
    find_specs = []
    for key_path_glob in path_helper.PathHelper.ExpandGlobStars(key_path, '\\'):
      logger.debug(f'building find spec from key path glob: {key_path_glob:s}')

      key_path_glob_upper = key_path_glob.upper()
      if key_path_glob_upper.startswith(
          'HKEY_LOCAL_MACHINE\\SYSTEM\\CURRENTCONTROLSET'):
        # Rewrite CurrentControlSet to ControlSet* for Windows NT.
        key_path_glob = ''.join([
            'HKEY_LOCAL_MACHINE\\System\\ControlSet*', key_path_glob[43:]])

      elif key_path_glob_upper.startswith('HKEY_USERS\\%%USERS.SID%%'):
        # Escaping charachter excluded from string index.
        key_path_glob = ''.join(['HKEY_CURRENT_USER', key_path_glob[24:]])

      find_spec = dfwinreg_registry_searcher.FindSpec(
          key_path_glob=key_path_glob)
      find_specs.append(find_spec)

    return find_specs

  def _BuildFindSpecsFromFileSourcePath(
      self, artifact_name, source_path, path_separator,
      environment_variables, user_accounts,
      enable_artifacts_map=False,
      original_registry_artifact_filter_names=None):
    """Builds find specifications from a file source type.

    Args:
      artifact_name (str): artifact name.
      source_path (str): file system path defined by the source.
      path_separator (str): file system path segment separator.
      environment_variables (list[EnvironmentVariableArtifact]): environment
          variables.
      user_accounts (list[UserAccountArtifact]): user accounts.
      enable_artifacts_map (Optional[bool]): True if the artifacts path map
          should be generated. Defaults to False.
      original_registry_artifact_filter_names (Optional[set[str]]): Set of
          original registry filter names, used in case Windows Registry hive
          files are being requested as a result of a previous filter.

    Returns:
      list[dfvfs.FindSpec]: find specifications for the file source type.
    """
    find_specs = []
    for path_glob in path_helper.PathHelper.ExpandGlobStars(
        source_path, path_separator):
      logger.debug(f'building find spec from path glob: {path_glob:s}')

      for path in path_helper.PathHelper.ExpandUsersVariablePath(
          path_glob, path_separator, user_accounts):
        logger.debug(f'building find spec from path: {path:s}')

        expanded_path = self._ExpandPathVariables(
            path, environment_variables, path_separator)
        if expanded_path is None:
          continue

        find_spec = self._CreateFindSpec(expanded_path, path_separator)
        if find_spec is None:
          continue

        find_specs.append(find_spec)

        if enable_artifacts_map:
          self._AddToArtifactsTrie(
              artifact_name, expanded_path,
              original_registry_artifact_filter_names,
              path_separator)

    return find_specs

  def _AddToArtifactsTrie(
      self, artifact_name, path,
      original_registry_artifact_filter_names,
      path_separator):
    """Adds a path to the artifacts trie.

    Args:
      artifact_name (str): artifact name.
      path (str): file system path.
      original_registry_artifact_filter_names (Optional[set[str]]): Set of
          original registry filter names.
      path_separator (str): path separator.
    """
    normalized_path = path.replace(path_separator, os.sep)
    self.artifacts_trie.AddPath(artifact_name, normalized_path, os.sep)
    if original_registry_artifact_filter_names:
      for name in original_registry_artifact_filter_names:
        self.artifacts_trie.AddPath(name, normalized_path, os.sep)

  def _ExpandPathVariables(self, path, environment_variables, path_separator):
    """Expands Windows paths and validates the result.

    Args:
      path (str): file system path with environment variables
      environment_variables (list[EnvironmentVariableArtifact]):
          environment variables.
      path_separator (str): file system path segment separator.

    Returns:
      str: expanded path, or None if the path is invalid.
    """
    if '%' in path:
      path = path_helper.PathHelper.ExpandWindowsPath(
          path, environment_variables)
      logger.debug(f'building find spec from expanded path: {path:s}')

    if not path.startswith(path_separator):
      logger.warning((
          f'The path filter must be defined as an absolute path: '
          f'"{path:s}"'))
      return None
    return path

  def _CreateFindSpec(self, path, path_separator):
    """Creates a dfVFS find specification.

    Args:
      path (str): Path to match.
      path_separator (str): file system path segment separator.

    Returns:
      dfvfs.FindSpec: a find specification or None if one cannot be created.
    """
    try:
      find_spec = dfvfs_file_system_searcher.FindSpec(
          case_sensitive=False, location_glob=path,
          location_separator=path_separator)
      return find_spec
    except ValueError as exception:
      logger.error((
          f'Unable to build find specification for path: "{path:s}" with '
          f'error: {exception!s}'))
      return None

  def BuildFindSpecs(
      self, artifact_filter_names, environment_variables=None,
      user_accounts=None, enable_artifacts_map=False,
      original_registry_artifact_filter_names=None):
    """Builds find specifications from artifact definitions.

    Args:
      artifact_filter_names (list[str]): names of artifact definitions that are
          used for filtering file system and Windows Registry key paths.
      environment_variables (list[EnvironmentVariableArtifact]): environment
          variables.
      user_accounts (Optional[list[UserAccountArtifact]]): user accounts.
      enable_artifacts_map (Optional[bool]): True if the artifacts path map
          should be generated. Defaults to False.
      original_registry_artifact_filter_names (Optional[set[str]]): Set of
          original registry filter names, used in case registry hive files
          are being requested as a result of a previous filter.
    """
    find_specs = {}
    for name in artifact_filter_names:
      definition = self._artifacts_registry.GetDefinitionByName(name)
      if not definition:
        definition = self._artifacts_registry.GetDefinitionByAlias(name)
      if not definition:
        logger.debug(f'undefined artifact definition: {name:s}')
        continue

      logger.debug(f'building find spec from artifact definition: {name:s}')
      artifact_find_specs = self._BuildFindSpecsFromArtifact(
          definition, environment_variables, user_accounts,
          enable_artifacts_map=enable_artifacts_map,
          original_registry_artifact_filter_names=(
              original_registry_artifact_filter_names))
      find_spec_list = find_specs.setdefault(name, [])
      find_spec_list.extend(artifact_find_specs)

    for name, find_spec_values in find_specs.items():
      for find_spec in find_spec_values:
        if isinstance(find_spec, dfvfs_file_system_searcher.FindSpec):
          self.file_system_find_specs.append(find_spec)

        elif isinstance(find_spec, dfwinreg_registry_searcher.FindSpec):
          self.registry_find_specs.append(find_spec)
          # Artifact names ordered similar to registry find specs.
          self.registry_find_specs_artifact_names.append(name)
        else:
          type_string = type(find_spec)
          logger.warning(
              f'Unsupported find specification type: {type_string!s}')

  @classmethod
  def CheckKeyCompatibility(cls, key_path):
    """Checks if a Windows Registry key path is supported by dfWinReg.

    Args:
      key_path (str): path of the Windows Registry key.

    Returns:
      bool: True if key is compatible or False if not.
    """
    key_path_upper = key_path.upper()
    for key_path_prefix in cls._COMPATIBLE_REGISTRY_KEY_PATH_PREFIXES:
      if key_path_upper.startswith(key_path_prefix):
        return True

    logger.warning(f'Key path: "{key_path:s}" is currently not supported')
    return False



"""Parser for Microsoft Edge load statistics database."""

from dfdatetime import webkit_time as dfdatetime_webkit_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class EdgeLoadStatisticsResourceEventData(events.EventData):
  """Microsoft Edge load statistics resource event data.

  Attributes:
    last_update: Last update time of resource, cached or not.
    query (str): query that created the event data.
    resource_hostname: External domain of the resource that was loaded
    resource_type: Integer descriptor of resource type
    top_level_hostname: Source domain that initiated resource load
  """

  DATA_TYPE = 'edge:resources:load_statistics'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.last_update = None
    self.query = None
    self.resource_hostname = None
    self.resource_type = None
    self.top_level_hostname = None


class EdgeLoadStatisticsPlugin(interface.SQLitePlugin):
  """SQLite parser plugin for Microsoft Edge load statistics database."""

  NAME = 'edge_load_statistics'
  DESCRIPTION = 'Parser for Microsoft Edge load_statistics.db'

  QUERIES = [
      ('SELECT top_level_hostname, resource_hostname, resource_type, '
       'last_update FROM load_statistics', 'ParseResourceRow')]

  REQUIRED_STRUCTURE = {
      'load_statistics': frozenset([
          'top_level_hostname', 'resource_hostname', 'resource_url_hash', 
          'resource_type', 'last_update']),
      'meta':frozenset([
          'key','value']),
      'redirect_statistics':frozenset([
          'source_hostname','destination_hostname',
          'is_top_level_document','last_update'])}

  SCHEMAS = [{
     'load_statistics': (
          'CREATE TABLE load_statistics(top_level_hostname TEXT,'
          'resource_hostname TEXT, resource_url_hash TEXT, resource_type'
          'INTEGER, last_update INTEGER NOT NULL,'
          'UNIQUE(top_level_hostname,resource_url_hash))'),
      'meta': (
          'CREATE TABLE meta(key LONGVARCHAR NOT NULL UNIQUE PRIMARY KEY,'
          'value LONGVARCHAR)'),
      'redirect_statistics': (
          'CREATE TABLE redirect_statistics(source_hostname TEXT,'
          'destination_hostname TEXT, is_top_level_document INTEGER NOT'
          'NULL, last_update INTEGER NOT NULL, UNIQUE(source_hostname,desti'
          'nation_hostname,is_top_level_document))')}]

  def _GetWebKitDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a WebKit date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.WebKitTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_webkit_time.WebKitTime(timestamp=timestamp)

  def ParseResourceRow(self, parser_mediator, query, row, **unused_kwargs):
    """Parses a row from the database.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row resulting from query.
    """
    # Note that pysqlite does not accept a Unicode string in row['string'] and
    # will raise "IndexError: Index must be int or string".
    query_hash = hash(query)

    event_data = EdgeLoadStatisticsResourceEventData()
    event_data.last_update = self._GetWebKitDateTimeRowValue(
        query_hash, row, 'last_update')
    event_data.query = query
    event_data.resource_hostname = self._GetRowValue(
        query_hash, row, 'resource_hostname')
    event_data.resource_type = self._GetRowValue(
        query_hash, row, 'resource_type')
    event_data.top_level_hostname = self._GetRowValue(
        query_hash, row, 'top_level_hostname')

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(EdgeLoadStatisticsPlugin)



"""Text parser plugin for Atlassian Bitbucket audit log files.

This is for the atlassian-bitbucket-audit.log file, one of multiple log files
produced by a Bitbucket DC/Server installation.

The audit log is a pipe-delimited file with the following fields:
  ip_address | event_name | user | timestamp_ms | entity | details |
  request_identifier | session_identifier

The timestamp field is milliseconds since the Unix epoch (January 1, 1970).

Also see:
  https://support.atlassian.com/bitbucket-data-center/kb/how-to-read-the-bitbucket-data-center-log-formats/
"""

import pyparsing

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class BitbucketAuditEventData(events.EventData):
  """Bitbucket audit log event data.

  Attributes:
    details (str): JSON-encoded details of the affected entity.
    entity (str): the affected entity in the format PRODUCT/key, such as
        BITBUCKET/bitbucket or PROJECT/myproject.
    event_name (str): the name of the audited event, such as
        RestrictedRefAddedEvent.
    recorded_time (dfdatetime.DateTimeValues): date and time the audit event
        was recorded.
    remote_address (str): remote IP address(es), including X-Forwarded-For
        proxies. Multiple addresses are comma-separated; the first address is
        the originating IP and the last is the directly connected IP.
    request_identifier (str): unique identifier for the request that triggered
        the audit event, correlatable with the access log.
    session_identifier (str): session identifier, correlatable with the access
        log.
    user_name (str): the name of the user who triggered the event.
  """

  DATA_TYPE = 'atlassian:bitbucket:audit'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.details = None
    self.entity = None
    self.event_name = None
    self.recorded_time = None
    self.remote_address = None
    self.request_identifier = None
    self.session_identifier = None
    self.user_name = None


class BitbucketAuditTextPlugin(interface.TextPlugin):
  """Text parser plugin for Atlassian Bitbucket audit log files."""

  NAME = 'bitbucket_audit'
  DATA_FORMAT = (
      'Atlassian Bitbucket audit log (atlassian-bitbucket-audit.log) file')

  ENCODING = 'utf-8'

  # Pipe as field separator.
  _SEPARATOR = pyparsing.Suppress(pyparsing.Literal('|'))

  # Remote IP address field: one or more IPv4/IPv6 addresses separated by
  # commas, such as "63.246.22.199,172.16.1.187" or "0:0:0:0:0:0:0:1"
  _REMOTE_ADDRESS = pyparsing.Combine(
      pyparsing.Word(pyparsing.alphanums + '.:,')).set_results_name(
          'remote_address')

  # Event name: CamelCase Java event class name, such as RestrictedRefAddedEvent
  _EVENT_NAME = pyparsing.Word(
      pyparsing.alphanums + '_').set_results_name('event_name')

  # User name: alphanumeric with dots/hyphens/underscores, or '-'
  _USER_NAME = (
      pyparsing.Word(pyparsing.alphanums + '-_./@') |
      pyparsing.Literal('-')).set_results_name('user_name')

  # Timestamp: milliseconds since Unix epoch, such as 1400681361906
  _TIMESTAMP_MS = pyparsing.Word(pyparsing.nums).set_parse_action(
      lambda tokens: int(tokens[0], 10)).set_results_name('timestamp_ms')

  # Entity: PRODUCT/key format, such as BITBUCKET/bitbucket or PROJECT/myproject
  # May also be '-' when not applicable
  _ENTITY = (
      pyparsing.Word(pyparsing.alphanums + '/-_.') |
      pyparsing.Literal('-')).set_results_name('entity')

  # Details: JSON object or array, or '-' when not applicable.
  # Use SkipTo to capture everything up to the next ' | ' separator.
  _DETAILS = pyparsing.SkipTo(
      pyparsing.Literal('|')).set_results_name('details')

  # Request identifier, such as '@8KJQAGx969x538x0', '*1APC3V1x...' or '-'.
  _REQUEST_IDENTIFIER = (
      pyparsing.Word(pyparsing.alphanums + '@*-_x') |
      pyparsing.Literal('-')).set_results_name('request_identifier')

  # Session identifier, short alphanumeric identifier or '-'.
  _SESSION_IDENTIFIER = (
      pyparsing.Word(pyparsing.alphanums + '-_') |
      pyparsing.Literal('-')).set_results_name('session_identifier')

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  # Audit log line format (pipe-delimited):
  # ip_address | event_name | user | timestamp_ms | entity | details |
  # request_identifier | session_identifier
  _AUDIT_LOG_LINE = (
      _REMOTE_ADDRESS + _SEPARATOR +
      _EVENT_NAME + _SEPARATOR +
      _USER_NAME + _SEPARATOR +
      _TIMESTAMP_MS + _SEPARATOR +
      _ENTITY + _SEPARATOR +
      _DETAILS + _SEPARATOR +
      _REQUEST_IDENTIFIER + _SEPARATOR +
      _SESSION_IDENTIFIER +
      _END_OF_LINE)

  _LINE_STRUCTURES = [('audit_entry', _AUDIT_LOG_LINE)]

  VERIFICATION_GRAMMAR = _AUDIT_LOG_LINE

  def _GetDateTimeValue(self, structure, name):
    """Retrieves a date and time value from a Pyparsing structure.

    Args:
      structure (pyparsing.ParseResults): tokens from a parsed log line.
      name (str): name of the token.

    Returns:
      dfdatetime.TimeElements: date and time value or None if not available.
    """
    timestamp = self._GetValueFromStructure(structure, name)
    if not timestamp:
      return None

    return dfdatetime_posix_time.PosixTimeInMilliseconds(
        timestamp=timestamp)

  def _GetStrippedValue(self, structure, name, default_value=None):
    """Retrieves a token value from a Pyparsing structure and strips '' or '-'.

    Args:
      structure (pyparsing.ParseResults): tokens from a parsed log line.
      name (str): name of the token.
      default_value (Optional[object]): default value.

    Returns:
      object: value in the token or default value if the token is not available
          in the structure.
    """
    value = self._GetValueFromStructure(structure, name)
    if not value or value == '-':
      return default_value

    return value

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    if key != 'audit_entry':
      raise errors.ParseError(f'Unsupported structure: {key:s}')

    details = self._GetValueFromStructure(
        structure, 'details', default_value='').strip()

    event_data = BitbucketAuditEventData()
    event_data.details = None if not details or details == '-' else details
    event_data.entity = self._GetStrippedValue(structure, 'entity')
    event_data.event_name = self._GetValueFromStructure(
        structure, 'event_name')
    event_data.recorded_time = self._GetDateTimeValue(structure, 'timestamp_ms')
    event_data.remote_address = self._GetValueFromStructure(
        structure, 'remote_address')
    event_data.request_identifier = self._GetStrippedValue(
        structure, 'request_identifier')
    event_data.session_identifier = self._GetStrippedValue(
        structure, 'session_identifier')
    event_data.user_name = self._GetStrippedValue(structure, 'user_name')

    parser_mediator.ProduceEventData(event_data)

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    timestamp_ms = self._GetValueFromStructure(structure, 'timestamp_ms')

    # The timestamp must look like a plausible millisecond epoch value.
    # Reject values that are too small (before year 2000) or too large.
    # Year 2000 in ms = 946684800000; year 2100 in ms = 4102444800000
    if (timestamp_ms is None or
        not 946684800000 <= timestamp_ms <= 4102444800000):
      return False

    return True


text_parser.TextLogParser.RegisterPlugin(BitbucketAuditTextPlugin)



"""Plist parser plugin for MacOS Airport plist files."""

from plaso.containers import events
from plaso.parsers import plist
from plaso.parsers.plist_plugins import interface


class MacOSAirportEventData(events.EventData):
  """MacOS airport event data.

  Attributes:
    last_connected_time (dfdatetime.DateTimeValues): last date and time MacOS
        Airport connected to the Wi-Fi network.
    security_type (str): Wi-Fi security type.
    ssid (str): Wi-Fi SSID.
  """

  DATA_TYPE = 'macos:airport:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.last_connected_time = None
    self.security_type = None
    self.ssid = None


class MacOSAirportPlistPlugin(interface.PlistPlugin):
  """Plist parser plugin for Airport plist files."""

  NAME = 'airport'
  DATA_FORMAT = 'Airport plist file'

  PLIST_PATH_FILTERS = frozenset([
      interface.PlistPathFilter('com.apple.airport.preferences.plist')])

  PLIST_KEYS = frozenset(['RememberedNetworks'])

  # pylint: disable=arguments-differ
  def _ParsePlist(self, parser_mediator, match=None, **unused_kwargs):
    """Extracts relevant Airport entries.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      match (Optional[dict[str: object]]): keys extracted from PLIST_KEYS.
    """
    for plist_key in match.get('RememberedNetworks', []):
      event_data = MacOSAirportEventData()
      event_data.last_connected_time = self._GetDateTimeValueFromPlistKey(
          plist_key, 'LastConnected')
      event_data.security_type = plist_key.get('SecurityType')
      event_data.ssid = plist_key.get('SSIDString')

      parser_mediator.ProduceEventData(event_data)


plist.PlistParser.RegisterPlugin(MacOSAirportPlistPlugin)



"""The storage reader."""

from plaso.containers import events
from plaso.containers import sessions
from plaso.storage import logger


class StorageReader:
  """Storage reader interface."""

  _CONTAINER_TYPE_SESSION = sessions.Session.CONTAINER_TYPE
  _CONTAINER_TYPE_EVENT_TAG = events.EventTag.CONTAINER_TYPE

  def __init__(self):
    """Initializes a storage reader."""
    super().__init__()
    self._serializers_profiler = None
    self._storage_profiler = None
    self._store = None

  def __enter__(self):
    """Make usable with "with" statement."""
    return self

  # pylint: disable=unused-argument
  def __exit__(self, exception_type, value, traceback):
    """Make usable with "with" statement."""
    self.Close()

  def Close(self):
    """Closes the storage reader."""
    self._store.Close()
    self._store = None

  def GetAttributeContainerByIdentifier(self, container_type, identifier):
    """Retrieves a specific type of container with a specific identifier.

    Args:
      container_type (str): container type.
      identifier (AttributeContainerIdentifier): attribute container identifier.

    Returns:
      AttributeContainer: attribute container or None if not available.
    """
    return self._store.GetAttributeContainerByIdentifier(
        container_type, identifier)

  def GetAttributeContainerByIndex(self, container_type, index):
    """Retrieves a specific attribute container.

    Args:
      container_type (str): attribute container type.
      index (int): attribute container index.

    Returns:
      AttributeContainer: attribute container or None if not available.
    """
    return self._store.GetAttributeContainerByIndex(container_type, index)

  def GetAttributeContainers(self, container_type, filter_expression=None):
    """Retrieves a specific type of attribute containers.

    Args:
      container_type (str): attribute container type.
      filter_expression (Optional[str]): expression to filter the resulting
          attribute containers by.

    Returns:
      generator(AttributeContainers): attribute container generator.
    """
    return self._store.GetAttributeContainers(
        container_type, filter_expression=filter_expression)

  def GetEventTagByEventIdentifer(self, event_identifier):
    """Retrieves the event tag of a specific event.

    Args:
      event_identifier (AttributeContainerIdentifier): event attribute
          container identifier.

    Returns:
      EventTag: event tag or None if the event has no event tag.
    """
    lookup_key = event_identifier.CopyToString()

    event_tags = list(self.GetAttributeContainers(
        self._CONTAINER_TYPE_EVENT_TAG,
        filter_expression=f'_event_identifier == "{lookup_key:s}"'))

    if not event_tags:
      return None

    if len(event_tags) > 1:
      logger.warning('More than 1 event tag returned.')

    return event_tags[0]

  def GetFormatVersion(self):
    """Retrieves the format version of the underlying storage file.

    Returns:
      int: the format version.
    """
    return self._store.format_version

  def GetNumberOfAttributeContainers(self, container_type):
    """Retrieves the number of a specific type of attribute containers.

    Args:
      container_type (str): attribute container type.

    Returns:
      int: the number of containers of a specified type.
    """
    return self._store.GetNumberOfAttributeContainers(container_type)

  def GetSerializationFormat(self):
    """Retrieves the serialization format of the underlying storage file.

    Returns:
      str: the serialization format.
    """
    return self._store.serialization_format

  def GetSessions(self):
    """Retrieves the sessions.

    Yields:
      Session: session attribute container.
    """
    yield from self.GetAttributeContainers(self._CONTAINER_TYPE_SESSION)

  def GetSortedEvents(self, time_range=None):
    """Retrieves the events in increasing chronological order.

    This includes all events written to the storage including those pending
    being flushed (written) to the storage.

    Args:
      time_range (Optional[TimeRange]): time range used to filter events
          that fall in a specific period.

    Returns:
      generator(EventObject): event generator.
    """
    return self._store.GetSortedEvents(time_range=time_range)

  def HasAttributeContainers(self, container_type):
    """Determines if a store contains a specific type of attribute container.

    Args:
      container_type (str): attribute container type.

    Returns:
      bool: True if the store contains the specified type of attribute
          containers.
    """
    return self._store.HasAttributeContainers(container_type)

  def SetSerializersProfiler(self, serializers_profiler):
    """Sets the serializers profiler.

    Args:
      serializers_profiler (SerializersProfiler): serializers profiler.
    """
    self._serializers_profiler = serializers_profiler
    if self._store:
      self._store.SetSerializersProfiler(serializers_profiler)

  def SetStorageProfiler(self, storage_profiler):
    """Sets the storage profiler.

    Args:
      storage_profiler (StorageProfiler): storage profiler.
    """
    self._storage_profiler = storage_profiler
    if self._store:
      self._store.SetStorageProfiler(storage_profiler)



"""Default plist parser plugin."""

import datetime

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import plist_event
from plaso.parsers import plist
from plaso.parsers.plist_plugins import interface


class DefaultPlugin(interface.PlistPlugin):
  """Default plist parser plugin."""

  NAME = 'plist_default'
  DATA_FORMAT = 'plist file'

  # pylint: disable=arguments-differ
  def _ParsePlist(self, parser_mediator, top_level=None, **unused_kwargs):
    """Extracts events from the values of entries within a plist.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      top_level (Optional[dict[str, object]]): plist top-level item.
    """
    for root, key_name, datetime_value in self._RecurseKey(top_level):
      if not isinstance(datetime_value, datetime.datetime):
        continue

      date_time = dfdatetime_time_elements.TimeElementsInMicroseconds()
      date_time.CopyFromDatetime(datetime_value)

      event_data = plist_event.PlistTimeEventData()
      event_data.key = key_name
      event_data.root = root
      event_data.written_time = date_time

      parser_mediator.ProduceEventData(event_data)

      # TODO: adjust code when there is a way to map keys to offsets.


plist.PlistParser.RegisterPlugin(DefaultPlugin)



"""Windows EventLog custom event formatter helpers."""

import re

from plaso.formatters import interface
from plaso.formatters import logger
from plaso.formatters import manager


class WindowsEventLogMessageFormatterHelper(
    interface.CustomEventFormatterHelper):
  """Windows EventLog message formatter helper."""

  IDENTIFIER = 'windows_eventlog_message'

  _PARAMETER_REGEX = re.compile(r'^%%[1-9][0-9]*$')

  def __init__(self):
    """Initialized a indows EventLog message formatter helper."""
    super().__init__()
    self._winevt_resources_helper = None

  def FormatEventValues(self, output_mediator, event_values):
    """Formats event values using the helper.

    Args:
      output_mediator (OutputMediator): output mediator.
      event_values (dict[str, object]): event values.
    """
    if not self._winevt_resources_helper:
      self._winevt_resources_helper = output_mediator.GetWinevtResourcesHelper()

    message_string = None
    provider_identifier = event_values.get('provider_identifier')
    source_name = event_values.get('source_name')
    message_identifier = event_values.get('message_identifier')
    event_version = event_values.get('event_version')
    if (provider_identifier or source_name) and message_identifier:
      message_string_template = self._winevt_resources_helper.GetMessageString(
          provider_identifier, source_name, message_identifier, event_version)
      if message_string_template:
        string_values = []
        for string_value in event_values.get('strings', []):
          if string_value is None:
            string_value = ''

          elif self._PARAMETER_REGEX.match(string_value):
            try:
              parameter_identifier = int(string_value[2:], 10)
              parameter_string = (
                  self._winevt_resources_helper.GetParameterString(
                      provider_identifier, source_name, parameter_identifier))
              if parameter_string:
                string_value = parameter_string

            except ValueError:
              pass

          string_values.append(string_value)

        try:
          message_string = message_string_template.format(*string_values)
        except (IndexError, TypeError) as exception:
          provider_identifier = provider_identifier or ''
          strings = ', '.join(string_values)
          logger.error((
              f'Unable to format message: 0x{message_identifier:08x} of '
              f'provider: {provider_identifier:s} template: '
              f'"{message_string_template:s}" and strings: "{strings:s}" '
              f'with error: {exception!s}'))
          # Unable to create the message string.
          # TODO: consider returning the unformatted message string.

    event_values['message_string'] = message_string


manager.FormattersManager.RegisterEventFormatterHelper(
    WindowsEventLogMessageFormatterHelper)



"""Plug-in to collect information about the Windows version."""

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.containers import events
from plaso.parsers import winreg_parser
from plaso.parsers.winreg_plugins import interface


class WindowsRegistryInstallationEventData(events.EventData):
  """Windows installation event data attribute container.

  Attributes:
    build_number (str): Windows build number.
    installation_time (dfdatetime.DateTimeValues): Windows installation date
        and time.
    key_path (str): Windows Registry key path.
    owner (str): registered owner.
    product_name (str): product name.
    service_pack (str): service pack.
    version (str): Windows version.
  """

  DATA_TYPE = 'windows:registry:installation'

  def __init__(self):
    """Initializes event data."""
    super().__init__(
        data_type=self.DATA_TYPE)
    self.build_number = None
    self.installation_time = None
    self.key_path = None
    self.owner = None
    self.product_name = None
    self.service_pack = None
    self.version = None


class WindowsVersionPlugin(interface.WindowsRegistryPlugin):
  """Plug-in to collect information about the Windows version."""

  NAME = 'windows_version'
  DATA_FORMAT = 'Windows version (product) Registry data'

  FILTERS = frozenset([
      interface.WindowsRegistryKeyPathFilter(
          'HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows NT\\'
          'CurrentVersion')])

  def ExtractEvents(self, parser_mediator, registry_key, **kwargs):
    """Extracts events from a Windows Registry key.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      registry_key (dfwinreg.WinRegistryKey): Windows Registry key.
    """
    # TODO: if not present indicate anomaly of missing InstallDate value.
    registry_value = registry_key.GetValueByName('InstallDate')
    if registry_value:
      event_data = WindowsRegistryInstallationEventData()
      event_data.build_number = self._GetValueFromKey(
          registry_key, 'CurrentBuildNumber')
      event_data.key_path = registry_key.path
      event_data.owner = self._GetValueFromKey(registry_key, 'RegisteredOwner')
      event_data.product_name = self._GetValueFromKey(
          registry_key, 'ProductName')
      event_data.service_pack = self._GetValueFromKey(
          registry_key, 'CSDVersion')
      event_data.version = self._GetValueFromKey(registry_key, 'CurrentVersion')

      installation_time = registry_value.GetDataAsObject()
      if installation_time:
        event_data.installation_time = dfdatetime_posix_time.PosixTime(
            timestamp=installation_time)

      parser_mediator.ProduceEventData(event_data)

    self._ProduceDefaultWindowsRegistryEvent(
        parser_mediator, registry_key, names_to_skip=['InstallDate'])


winreg_parser.WinRegistryParser.RegisterPlugin(WindowsVersionPlugin)



"""Windows event data attribute containers."""

from dfdatetime import uuid_time as dfdatetime_uuid_time

from plaso.containers import events


class WindowsDistributedLinkTrackingEventData(events.EventData):
  """Windows distributed link event data attribute container.

  Attributes:
    creation_time (dfdatetime.DateTimeValues): file entry creation date
        and time.
    mac_address (str): MAC address stored in the UUID.
    origin (str): origin of the event (event source).
        E.g. the path of the corresponding LNK file or file reference
        MFT entry with the corresponding NTFS $OBJECT_ID attribute.
    uuid (str): UUID.
  """

  DATA_TYPE = 'windows:distributed_link_tracking:creation'

  def __init__(self, uuid, origin):
    """Initializes an event object.

    Args:
      uuid (uuid.UUID): UUID.
      origin (str): origin of the event (event source).
          E.g. the path of the corresponding LNK file or file reference
          MFT entry with the corresponding NTFS $OBJECT_ID attribute.

    Raises:
      ValueError: if the UUID version is not supported.
    """
    if uuid.version != 1:
      raise ValueError('Unsupported UUID version.')

    mac_address = (f'{uuid.hex[20:22]:s}:{uuid.hex[22:24]:s}:'
                   f'{uuid.hex[24:26]:s}:{uuid.hex[26:28]:s}:'
                   f'{uuid.hex[28:30]:s}:{uuid.hex[30:32]:s}')

    super().__init__(
        data_type=self.DATA_TYPE)
    self.creation_time = dfdatetime_uuid_time.UUIDTime(timestamp=uuid.time)
    self.mac_address = mac_address
    # TODO: replace origin my something machine readable.
    self.origin = origin
    self.uuid = f'{uuid!s}'


class WindowsRegistryEventData(events.EventData):
  """Windows Registry event data attribute container.

  Attributes:
    key_path (str): Windows Registry key path.
    last_written_time (dfdatetime.DateTimeValues): key last written date and
        time.
    values (list[tuple[str, str, str]]): name, data type and data of the values
        in the key.
  """

  DATA_TYPE = 'windows:registry:key_value'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.key_path = None
    self.last_written_time = None
    self.values = None


class WindowsShellItemFileEntryEventData(events.EventData):
  """Windows shell item file entry event data attribute container.

  Attributes:
    access_time (dfdatetime.DateTimeValues): file entry last access date
        and time.
    creation_time (dfdatetime.DateTimeValues): file entry creation date
        and time.
    file_reference (str): NTFS file reference, in the format:
        "MTF entry - sequence number".
    localized_name (str): localized name of the file entry shell item.
    long_name (str): long name of the file entry shell item.
    modification_time (dfdatetime.DateTimeValues): file entry last modification
        date and time.
    name (str): name of the file entry shell item.
    origin (str): origin of the event.
    shell_item_path (str): shell item path.
  """

  DATA_TYPE = 'windows:shell_item:file_entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(
        data_type=self.DATA_TYPE)
    self.access_time = None
    self.creation_time = None
    self.file_reference = None
    self.localized_name = None
    self.long_name = None
    self.modification_time = None
    self.name = None
    self.origin = None
    self.shell_item_path = None


class WindowsVolumeEventData(events.EventData):
  """Windows volume event data attribute container.

  Attributes:
    creation_time (dfdatetime.DateTimeValues): volume creation date and time.
    device_path (str): volume device path.
    origin (str): origin of the event (event source), for example
        the corresponding Prefetch file name.
    serial_number (str): volume serial number.
  """

  DATA_TYPE = 'windows:volume:creation'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.creation_time = None
    self.device_path = None
    # TODO: replace origin with something machine readable.
    self.origin = None
    self.serial_number = None



"""YAML-based timeliner configuration file."""

import yaml

from plaso.lib import errors


class TimelinerDefinition:
  """Timeliner definition.

  Attributes:
    attribute_mappings (dict[str, str]): date and time description
        (timestamp_desc) per attribute name.
    data_type (str): event data type indicator.
    place_holder_event (bool): True if the timeliner should generate
        a placeholder event if no date and time values were found in
        the event data.
  """

  def __init__(self, data_type):
    """Initializes a timeliner definition.

    Args:
      data_type (str): event data type indicator.
    """
    super().__init__()
    self.attribute_mappings = {}
    self.data_type = data_type
    self.place_holder_event = True


class YAMLTimelinerConfigurationFile:
  """YAML-based timeliner configuration file.

  A YAML-based timeliner configuration file contains one or more timeliner
  definitions. A timeliner definitions consists of:

  data_type: 'fs:stat'
  attribute_mappings:
  - name: 'access_time'
    description: 'Last Access Time'
  place_holder_event: true

  Where:
  * data_type, defines the corresponding event data type;
  * attribute_mappings, defines attribute mappings;
  * place_holder_event, defines if the timeliner should generate a placeholder
    event.
  """

  _SUPPORTED_KEYS = frozenset([
      'attribute_mappings',
      'data_type',
      'place_holder_event'])

  def _ReadTimelinerDefinition(self, timeliner_definition_values):
    """Reads a timeliner definition from a dictionary.

    Args:
      timeliner_definition_values (dict[str, object]): timeliner definition
           values.

    Returns:
      TimelinerDefinition: a timeliner definition.

    Raises:
      ParseError: if the format of the timeliner definition is not set
          or incorrect.
    """
    if not timeliner_definition_values:
      raise errors.ParseError('Missing timeliner definition values.')

    different_keys = set(timeliner_definition_values) - self._SUPPORTED_KEYS
    if different_keys:
      different_keys = ', '.join(different_keys)
      raise errors.ParseError(f'Undefined keys: {different_keys:s}')

    data_type = timeliner_definition_values.get('data_type')
    if not data_type:
      raise errors.ParseError(
          'Invalid event timeliner definition missing data type.')

    attribute_mappings = timeliner_definition_values.get(
        'attribute_mappings', None) or []

    timeliner_definition = TimelinerDefinition(data_type)
    timeliner_definition.attribute_mappings = {
        attribute_mapping['name']: attribute_mapping['description']
        for attribute_mapping in attribute_mappings}
    timeliner_definition.place_holder_event = timeliner_definition_values.get(
        'place_holder_event', True)

    return timeliner_definition

  def _ReadFromFileObject(self, file_object):
    """Reads the timeliner definitions from a file-like object.

    Args:
      file_object (file): timeliner definitions file-like object.

    Yields:
      TimelinerDefinition: a timeliner definition.
    """
    yaml_generator = yaml.safe_load_all(file_object)

    for yaml_definition in yaml_generator:
      yield self._ReadTimelinerDefinition(yaml_definition)

  def ReadFromFile(self, path):
    """Reads the timeliner definitions from a YAML file.

    Args:
      path (str): path to a timeliner configuration file.

    Yields:
      TimelinerDefinition: a timeliner definition.
    """
    with open(path, 'r', encoding='utf-8') as file_object:
      yield from self._ReadFromFileObject(file_object)



"""SQLite parser plugin for IMO HD chat message database files."""

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class IMOHDChatMessageEventData(events.EventData):
  """IMO HD chat message event data.

  Attributes:
    alias (str): alias (or name) of the author of the message.
    query (str): SQL query that was used to obtain the event data.
    recorded_time (dfdatetime.DateTimeValues): date and time the chat
        message was recorded.
    status (int): send status.
    text (str): text of the chat message.
  """

  DATA_TYPE = 'ios:imohdchat:message'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.alias = None
    self.query = None
    self.recorded_time = None
    self.status = None
    self.text = None


class IMOHDChatMessagePlugin(interface.SQLitePlugin):
  """SQLite parser plugin for IMO HD chat message database files."""

  NAME = 'ios_imohdchat_message'
  DATA_FORMAT = (
      'IMO HD chat message SQLite database (IMODb2.sqlite) file')

  REQUIRED_STRUCTURE = {
      'ZIMOCHATMSG': frozenset([
          'Z_PK', 'ZALIAS', 'ZTEXT', 'ZISSENT', 'ZTS'])}

  QUERIES = [(
      'SELECT ZTS, ZALIAS, ZTEXT, ZISSENT FROM ZIMOCHATMSG',
      'ParseApplicationUsageRow')]

  def _GetDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.PosixTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_posix_time.PosixTimeInNanoseconds(timestamp=timestamp)

  def ParseApplicationUsageRow(
      self, parser_mediator, query, row, **unused_kwargs):
    """Parses an application usage row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    # TODO: add attribute for ZSENDERTS
    event_data = IMOHDChatMessageEventData()
    event_data.alias = self._GetRowValue(query_hash, row, 'ZALIAS')
    event_data.query = query
    event_data.recorded_time = self._GetDateTimeRowValue(query_hash, row, 'ZTS')
    event_data.status = self._GetRowValue(query_hash, row, 'ZISSENT')
    event_data.text = self._GetRowValue(query_hash, row, 'ZTEXT')

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(IMOHDChatMessagePlugin)



"""Text parser plugin for Sophos anti-virus logs (SAV.txt) files."""

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class SophosAVLogEventData(events.EventData):
  """Sophos anti-virus log event data.

  Attributes:
    added_time (dfdatetime.DateTimeValues): date and time the log entry
        was added.
    text (str): Sophos anti-virus log message.
  """

  DATA_TYPE = 'sophos:av:log'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.added_time = None
    self.text = None


class SophosAVLogTextPlugin(interface.TextPlugin):
  """Text parser plugin for Sophos anti-virus logs (SAV.txt) files."""

  NAME = 'sophos_av'
  DATA_FORMAT = 'Sophos anti-virus log file (SAV.txt) file'

  ENCODING = 'utf-16-le'

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  # Date and time values are formatted as: YYYYMMDD hhmmss
  # For example: 20100720 183814
  _DATE_TIME = pyparsing.Group(
      _FOUR_DIGITS + _TWO_DIGITS + _TWO_DIGITS +
      _TWO_DIGITS + _TWO_DIGITS + _TWO_DIGITS)

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  _LOG_LINE = (
      _DATE_TIME.set_results_name('date_time') +
      pyparsing.restOfLine().set_results_name('text') +
      _END_OF_LINE)

  _LINE_STRUCTURES = [('log_line', _LOG_LINE)]

  VERIFICATION_GRAMMAR = _LOG_LINE

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    event_data = SophosAVLogEventData()
    event_data.added_time = self._ParseTimeElements(time_elements_structure)
    event_data.text = self._GetStringValueFromStructure(structure, 'text')

    parser_mediator.ProduceEventData(event_data)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
          of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      # Ensure time_elements_tuple is not a pyparsing.ParseResults otherwise
      # copy.deepcopy() of the dfDateTime object will fail on Python 3.8 with:
      # "TypeError: 'str' object is not callable" due to pyparsing.ParseResults
      # overriding __getattr__ with a function that returns an empty string
      # when named token does not exists.

      year, month, day_of_month, hours, minutes, seconds = (
          time_elements_structure)

      time_elements_tuple = (year, month, day_of_month, hours, minutes, seconds)

      date_time = dfdatetime_time_elements.TimeElements(
          time_elements_tuple=time_elements_tuple)

      # TODO: check if date and time values are local time or in UTC.
      date_time.is_local_time = True

      return date_time

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    try:
      self._ParseTimeElements(time_elements_structure)
    except ValueError:
      return False

    return True


text_parser.TextLogParser.RegisterPlugin(SophosAVLogTextPlugin)



"""SQLite parser plugin for Mozilla Firefox cookies database files."""

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.containers import events
from plaso.lib import cookie_plugins_helper
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class FirefoxCookieEventData(events.EventData):
  """Firefox Cookie event data.

  Attributes:
    access_time (dfdatetime.DateTimeValues): date and time the cookie
        was last accessed.
    cookie_name (str): name field of the cookie.
    creation_time (dfdatetime.DateTimeValues): date and time the cookie
        was created.
    data (str): cookie data.
    expiration_time (dfdatetime.DateTimeValues): date and time the cookie
        expires.
    httponly (bool): True if the cookie cannot be accessed through client
        side script.
    host (str): hostname of host that set the cookie value.
    offset (str): identifier of the row, from which the event data was
        extracted.
    path (str): URI of the page that set the cookie.
    query (str): SQL query that was used to obtain the event data.
    secure (bool): True if the cookie should only be transmitted over a secure
        channel.
  """

  DATA_TYPE = 'firefox:cookie:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.access_time = None
    self.cookie_name = None
    self.creation_time = None
    self.data = None
    self.expiration_time = None
    self.host = None
    self.httponly = None
    self.offset = None
    self.path = None
    self.query = None
    self.secure = None
    self.url = None


class BaseFirefoxCookiePlugin(
    interface.SQLitePlugin, cookie_plugins_helper.CookiePluginsHelper):
  """Shared SQLite parser plugin for Mozilla Firefox cookies database files."""

  def _GetPosixTimeDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a POSIX time (in seconds) date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.PosixTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    # Note that pysqlite3 can return an empty string for a NULL value.
    # Also see: https://github.com/log2timeline/plaso/issues/4961
    if timestamp is None or not isinstance(timestamp, int):
      return None

    return dfdatetime_posix_time.PosixTime(timestamp=timestamp)

  def _GetPosixTimeInMicrosecondsDateTimeRowValue(
      self, query_hash, row, value_name):
    """Retrieves a POSIX time in microseconds date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.PosixTimeInMicroseconds: date and time value or None if not
          available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    # Note that pysqlite3 can return an empty string for a NULL value.
    # Also see: https://github.com/log2timeline/plaso/issues/4961
    if timestamp is None or not isinstance(timestamp, int):
      return None

    return dfdatetime_posix_time.PosixTimeInMicroseconds(timestamp=timestamp)

  def ParseCookieRow(self, parser_mediator, query, row, **unused_kwargs):
    """Parses a cookie row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    cookie_data = self._GetRowValue(query_hash, row, 'value')
    cookie_name = self._GetRowValue(query_hash, row, 'name')

    hostname = self._GetRowValue(query_hash, row, 'host')
    if hostname.startswith('.'):
      hostname = hostname[1:]

    is_secure = bool(self._GetRowValue(query_hash, row, 'isSecure'))
    if is_secure:
      url_scheme = 'https'
    else:
      url_scheme = 'http'

    path = self._GetRowValue(query_hash, row, 'path')
    url = f'{url_scheme:s}://{hostname:s}{path:s}'

    event_data = FirefoxCookieEventData()
    event_data.access_time = self._GetPosixTimeInMicrosecondsDateTimeRowValue(
        query_hash, row, 'lastAccessed')
    event_data.cookie_name = cookie_name
    event_data.creation_time = self._GetPosixTimeInMicrosecondsDateTimeRowValue(
        query_hash, row, 'creationTime')
    event_data.data = cookie_data
    event_data.host = hostname
    event_data.httponly = bool(self._GetRowValue(query_hash, row, 'isHttpOnly'))
    event_data.offset = self._GetRowValue(query_hash, row, 'id')
    event_data.path = path
    event_data.query = query
    event_data.secure = is_secure
    event_data.url = url

    # Expiry time (nsCookieService::GetExpiry in
    # netwerk/cookie/nsCookieService.cpp).
    # It's calculated as the difference between the server time and the time
    # the server wants the cookie to expire and adding that difference to the
    # client time. This localizes the client time regardless of whether or not
    # the TZ environment variable was set on the client.

    event_data.expiration_time = self._GetPosixTimeDateTimeRowValue(
        query_hash, row, 'expiry')

    parser_mediator.ProduceEventData(event_data)

    self._ParseCookie(parser_mediator, cookie_name, cookie_data, url)


class FirefoxCookie2Plugin(BaseFirefoxCookiePlugin):
  """SQLite parser plugin for Mozilla Firefox cookies schema 2 databases.

  Also see:
    https://hg-edge.mozilla.org/mozilla-central/file/349a2f003529/netwerk/cookie/nsCookie.h
  """

  NAME = 'firefox_2_cookies'
  DATA_FORMAT = 'Mozilla Firefox cookies SQLite database file version 2'

  REQUIRED_STRUCTURE = {
      'moz_cookies': frozenset([
          'id', 'baseDomain', 'name', 'value', 'host', 'path', 'expiry',
          'lastAccessed', 'creationTime', 'isSecure', 'isHttpOnly'])}

  QUERIES = [
      (('SELECT id, baseDomain, name, value, host, path, expiry, '
        'lastAccessed, creationTime, isSecure, isHttpOnly FROM moz_cookies'),
       'ParseCookieRow')]

  SCHEMAS = [{
      'moz_cookies': (
          'CREATE TABLE moz_cookies (id INTEGER PRIMARY KEY, baseDomain TEXT, '
          'appId INTEGER DEFAULT 0, inBrowserElement INTEGER DEFAULT 0, name '
          'TEXT, value TEXT, host TEXT, path TEXT, expiry INTEGER, '
          'lastAccessed INTEGER, creationTime INTEGER, isSecure INTEGER, '
          'isHttpOnly INTEGER, CONSTRAINT moz_uniqueid UNIQUE (name, host, '
          'path, appId, inBrowserElement))')}]


class FirefoxCookie10Plugin(BaseFirefoxCookiePlugin):
  """SQLite parser plugin for Mozilla Firefox cookies schema 10 databases.

  In schema 10 baseDomain was removed.
  """

  NAME = 'firefox_10_cookies'
  DATA_FORMAT = 'Mozilla Firefox cookies SQLite database file version 10'

  REQUIRED_STRUCTURE = {
      'moz_cookies': frozenset([
          'id', 'name', 'value', 'host', 'path', 'expiry',
          'lastAccessed', 'creationTime', 'isSecure', 'isHttpOnly'])}

  QUERIES = [
      (('SELECT id, name, value, host, path, expiry, '
        'lastAccessed, creationTime, isSecure, isHttpOnly FROM moz_cookies'),
       'ParseCookieRow')]

  SCHEMAS = [{
      'moz_cookies': (
          'CREATE TABLE moz_cookies (id INTEGER PRIMARY KEY, '
          'appId INTEGER DEFAULT 0, inBrowserElement INTEGER DEFAULT 0, name '
          'TEXT, value TEXT, host TEXT, path TEXT, expiry INTEGER, '
          'lastAccessed INTEGER, creationTime INTEGER, isSecure INTEGER, '
          'isHttpOnly INTEGER, CONSTRAINT moz_uniqueid UNIQUE (name, host, '
          'path, appId, inBrowserElement))')}]


sqlite.SQLiteParser.RegisterPlugins([
    FirefoxCookie2Plugin, FirefoxCookie10Plugin])



"""SQLite parser plugin for Dropbox sync_history database files."""

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class DropboxSyncHistoryEventData(events.EventData):
  """Dropbox Sync History Database event data.

  Attributes:
    event_type (str): the event type
    file_event_type (str): the file event type
    direction (str): the source of the synchronisation event
    file_identifier (str): the Dropbox identifier of the file.
    local_path (str): the local path of the file.
    recorded_time (dfdatetime.DateTimeValues): date and time the log entry
        was recorded.
  """

  DATA_TYPE = 'dropbox:sync_history:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.event_type = None
    self.file_event_type = None
    self.direction = None
    self.file_identifier = None
    self.local_path = None
    self.recorded_time = None


class DropboxSyncDatabasePlugin(interface.SQLitePlugin):
  """SQLite parser plugin for Dropbox sync_history.db database files.

  The Linux sync_history.db database is typically stored in:
  $HOME/.dropbox/instance1/sync_history.db

  The Windows 10 sync_history.db database is typically stored in:
  $HOME/AppData/Local/Dropbox/instance1/sync_history.db
  """

  NAME = 'dropbox'
  DATA_FORMAT = 'Dropbox sync history database (sync_history.db) file'

  REQUIRED_STRUCTURE = {
      'sync_history': frozenset(['timestamp', 'event_type', 'file_event_type',
      'direction', 'file_id', 'local_path'])}

  QUERIES = [
      ('SELECT timestamp, event_type, file_event_type, direction, file_id, '
       'local_path FROM sync_history;',
       'ParseSyncHistoryRow')]

  SCHEMAS = [{
      'sync_history': (
          'CREATE TABLE sync_history (event_type TEXT NOT NULL, file_event_type'
          ' TEXT, direction TEXT, file_id TEXT, local_path TEXT, timestamp '
          'INTEGER NOT NULL, other_user INTEGER')}]

  def _GetDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.PosixTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_posix_time.PosixTime(timestamp=timestamp)

  def ParseSyncHistoryRow(self, parser_mediator, query, row, **unused_kwargs):
    """Parses a sync_history row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    event_data = DropboxSyncHistoryEventData()
    event_data.event_type = self._GetRowValue(query_hash, row, 'event_type')
    event_data.file_event_type = self._GetRowValue(
        query_hash, row, 'file_event_type')
    event_data.direction = self._GetRowValue(query_hash, row, 'direction')
    event_data.file_identifier = self._GetRowValue(query_hash, row, 'file_id')
    event_data.local_path = self._GetRowValue(query_hash, row, 'local_path')
    event_data.recorded_time = self._GetDateTimeRowValue(
        query_hash, row, 'timestamp')

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(DropboxSyncDatabasePlugin)



"""Text parser plugin for PowerShell transcript log files."""

import copy
import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class PowerShellTranscriptLogEventData(events.EventData):
  """PowerShell transcript log event data.

  Attributes:
    build_version (str): Build number of current version.
    clr_version (str): Common Language Runtime version.
    commands (str): Commands that were executed.
    compatible_versions (str): Compatible PowerShell versions.
    configuration_name (str): Configuration name.
    edition (str): PowerShell edition
    host_application (str): Application that executed the commands.
    machine (str): Hostname of machine.
    process_identifier (str): Process identifier.
    remoting_protocol_version (str): PowerShell remote management protocol
        version.
    runas_user (str): User context of execution.
    serialization_version (str): Serialization method version.
    start_time (dfdatetime.DateTimeValues): date and time the start of
        the PowerShell transcript.
    username (str): User that executed the commands.
    version (str): PowerShell version.
    ws_man_stack_version (str): WS-Management stack version
  """

  DATA_TYPE = 'powershell:transcript_log:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(
        data_type=self.DATA_TYPE)
    self.build_version = None
    self.clr_version = None
    self.commands = None
    self.compatible_versions = None
    self.configuration_name = None
    self.edition = None
    self.host_application = None
    self.machine = None
    self.process_identifier = None
    self.remoting_protocol_version = None
    self.runas_user = None
    self.serialization_version = None
    self.start_time = None
    self.username = None
    self.version = None
    self.ws_man_stack_version = None


class PowerShellTranscriptLogTextPlugin(interface.TextPlugin):
  """Text parser plugin for PowerShell transcript log files."""

  NAME = 'powershell_transcript'

  DATA_FORMAT = 'PowerShell transcript event'

  ENCODING = 'utf-8'

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  # Date and time values are formatted as: YYYYMMDDhhmmss
  # For example: 20220824124237
  _DATE_TIME = (_FOUR_DIGITS + _TWO_DIGITS + _TWO_DIGITS +
                _TWO_DIGITS + _TWO_DIGITS + _TWO_DIGITS)

  _SEPARATOR = pyparsing.Literal('**********************')

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  _SEPARATOR_LINE = _SEPARATOR + _END_OF_LINE

  _TRANSSCRIPT_START_LINE = pyparsing.Regex(r'.*Windows PowerShell.*\n')

  # A Metadata key always start with an uppercase character.
  _METADATA_KEY = pyparsing.Word(
      pyparsing.alphas.upper(), pyparsing.alphas + '- ')

  _METADATA_LINE = pyparsing.Combine(
      pyparsing.Suppress(_METADATA_KEY) + pyparsing.Suppress(':') +
      pyparsing.restOfLine() + _END_OF_LINE)

  _LOG_LINE = (
      pyparsing.NotAny(_SEPARATOR) +
      pyparsing.restOfLine().set_results_name('body') + _END_OF_LINE)

  _HEADER_GRAMMAR = (
      _SEPARATOR_LINE + _TRANSSCRIPT_START_LINE +
      _METADATA_LINE.set_results_name('date_time') +
      _METADATA_LINE.set_results_name('username') +
      _METADATA_LINE.set_results_name('runas_user') +
      _METADATA_LINE.set_results_name('configuration_name') +
      _METADATA_LINE.set_results_name('machine') +
      _METADATA_LINE.set_results_name('host_application') +
      _METADATA_LINE.set_results_name('process_identifier') +
      _METADATA_LINE.set_results_name('version') +
      _METADATA_LINE.set_results_name('edition') +
      _METADATA_LINE.set_results_name('compatible_versions') +
      _METADATA_LINE.set_results_name('build_version') +
      _METADATA_LINE.set_results_name('clr_version') +
      _METADATA_LINE.set_results_name('ws_man_stack_version') +
      _METADATA_LINE.set_results_name('remoting_protocol_version') +
      _METADATA_LINE.set_results_name('serialization_version') +
      _SEPARATOR_LINE)

  _LINE_STRUCTURES = [
      ('log_line', _LOG_LINE),
      ('separator_line', _SEPARATOR_LINE)]

  VERIFICATION_GRAMMAR = _SEPARATOR_LINE + _TRANSSCRIPT_START_LINE

  VERIFICATION_LITERALS = ['Windows PowerShell']

  # TODO: handle footer with end time.

  def __init__(self):
    """Initializes a text parser plugin."""
    super().__init__()
    self._command_history = []
    self._event_data = None
    self._in_command_history = False

  def _ParseHeader(self, parser_mediator, text_reader):
    """Parses a text-log file header.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Raises:
      ParseError: when the header cannot be parsed.
    """
    try:
      structure_generator = self._HEADER_GRAMMAR.scan_string(
          text_reader.lines, max_matches=1)
      structure, start, end = next(structure_generator)

    except StopIteration:
      structure = None

    except pyparsing.ParseException as exception:
      raise errors.ParseError(exception)

    if not structure or start != 0:
      raise errors.ParseError('No match found.')

    event_data = PowerShellTranscriptLogEventData()
    event_data.build_version = self._GetStringValueFromStructure(
        structure, 'build_version')
    event_data.clr_version = self._GetStringValueFromStructure(
        structure, 'clr_version')
    event_data.compatible_versions = self._GetStringValueFromStructure(
        structure, 'compatible_versions')
    event_data.configuration_name = self._GetStringValueFromStructure(
        structure, 'configuration_name')
    event_data.edition = self._GetStringValueFromStructure(
        structure, 'edition')
    event_data.host_application = self._GetStringValueFromStructure(
        structure, 'host_application')
    event_data.machine = self._GetStringValueFromStructure(
        structure, 'machine')
    event_data.process_identifier = self._GetStringValueFromStructure(
        structure, 'process_identifier')
    event_data.remoting_protocol_version = self._GetStringValueFromStructure(
        structure, 'remoting_protocol_version')
    event_data.runas_user = self._GetStringValueFromStructure(
        structure, 'runas_user')
    event_data.serialization_version = self._GetStringValueFromStructure(
        structure, 'serialization_version')
    event_data.username = self._GetStringValueFromStructure(
        structure, 'username')
    event_data.version = self._GetStringValueFromStructure(
        structure, 'version')
    event_data.ws_man_stack_version = self._GetStringValueFromStructure(
        structure, 'ws_man_stack_version')

    date_time_structure = self._GetStringValueFromStructure(
        structure, 'date_time')

    try:
      time_elements_structure = self._DATE_TIME.parse_string(
          date_time_structure)
    except pyparsing.ParseException:
      raise errors.ParseError('Unable to parse date time.')

    event_data.start_time = self._ParseTimeElements(time_elements_structure)

    self._event_data = event_data
    self._in_command_history = True

    text_reader.SkipAhead(end)

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    if self._in_command_history:
      if key == 'log_line':
        body = self._GetStringValueFromStructure(structure, 'body')
        if body:
          self._command_history.append(body)

      elif key == 'separator_line':
        event_data = copy.deepcopy(self._event_data)

        event_data.commands = '; '.join(self._command_history)

        parser_mediator.ProduceEventData(event_data)

        self._command_history = []
        self._in_command_history = False

    else:
      if key == 'log_line':
        body = self._GetStringValueFromStructure(structure, 'body')
        if ':' in body:
          date_time_structure = body.rsplit(':', maxsplit=1)[-1].strip()

          try:
            time_elements_structure = self._DATE_TIME.parse_string(
                date_time_structure)
          except pyparsing.ParseException:
            raise errors.ParseError('Unable to parse date time.')

          self._event_data.start_time = self._ParseTimeElements(
              time_elements_structure)

      elif key == 'separator_line':
        self._in_command_history = True

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
          of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      # Ensure time_elements_tuple is not a pyparsing.ParseResults otherwise
      # copy.deepcopy() of the dfDateTime object will fail on Python 3.8 with:
      # "TypeError: 'str' object is not callable" due to pyparsing.ParseResults
      # overriding __getattr__ with a function that returns an empty string
      # when named token does not exists.

      year, month, day_of_month, hours, minutes, seconds = (
          time_elements_structure)

      time_elements_tuple = (year, month, day_of_month, hours, minutes, seconds)

      date_time = dfdatetime_time_elements.TimeElements(
          time_elements_tuple=time_elements_tuple)

      date_time.is_local_time = True

      return date_time

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def _ResetState(self):
    """Resets stored values."""
    self._command_history = []
    self._event_data = None
    self._in_command_history = False

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    self._ResetState()

    return True


text_parser.TextLogParser.RegisterPlugin(PowerShellTranscriptLogTextPlugin)



"""Count related attribute container definitions."""

from acstore.containers import interface
from acstore.containers import manager


class EventLabelCount(interface.AttributeContainer):
  """Event label count attribute container.

  Attributes:
    label (str): event label.
    number_of_events (int): number of events with label.
  """

  CONTAINER_TYPE = 'event_label_count'

  SCHEMA = {
      'label': 'str',
      'number_of_events': 'int'}

  def __init__(self, label=None, number_of_events=None):
    """Initializes an event label count attribute container.

    Args:
      label (Optional[str]): event label.
      number_of_events (Optional[int]): number of events with label.
          the parser or parser plugin.
    """
    super().__init__()
    self.label = label
    self.number_of_events = number_of_events


class ParserCount(interface.AttributeContainer):
  """Parser count attribute container.

  Attributes:
    name (str): name of the parser or parser plugin.
    number_of_events (int): number of events generated by the parser or
        parser plugin.
  """

  CONTAINER_TYPE = 'parser_count'

  SCHEMA = {
      'name': 'str',
      'number_of_events': 'int'}

  def __init__(self, name=None, number_of_events=None):
    """Initializes a parser count attribute container.

    Args:
      name (Optional[str]): name of the parser or parser plugin.
      number_of_events (Optional[int]): number of events generated by
          the parser or parser plugin.
    """
    super().__init__()
    self.name = name
    self.number_of_events = number_of_events


manager.AttributeContainersManager.RegisterAttributeContainers([
    EventLabelCount, ParserCount])



"""MacOS preprocessor plugins."""

import abc
import plistlib

from plaso.containers import artifacts
from plaso.lib import errors
from plaso.lib import plist
from plaso.preprocessors import interface
from plaso.preprocessors import manager


class PlistFileArtifactPreprocessorPlugin(
    interface.FileArtifactPreprocessorPlugin):
  """Plist file artifact preprocessor plugin interface.

  Retrieves values from a plist file artifact using names of keys defined in
  _PLIST_KEYS.
  """

  # The key that's value should be returned back. It is an ordered list
  # of preference. If the first value is found it will be returned and no
  # others will be searched.
  _PLIST_KEYS = ['']

  def _FindKeys(self, key, names, matches):
    """Searches the plist key hierarchy for keys with matching names.

    If a match is found a tuple of the key name and value is added to
    the matches list.

    Args:
      key (dict[str, object]): plist key.
      names (list[str]): names of the keys to match.
      matches (list[str]): keys with matching names.
    """
    for name, subkey in key.items():
      if name in names:
        matches.append((name, subkey))

      if isinstance(subkey, dict):
        self._FindKeys(subkey, names, matches)

  def _ParseFileData(self, mediator, file_object):
    """Parses file content (data) for a preprocessing attribute.

    Args:
      mediator (PreprocessMediator): mediates interactions between preprocess
          plugins and other components, such as storage.
      file_object (dfvfs.FileIO): file-like object that contains the artifact
          value data.

    Raises:
      errors.PreProcessFail: if the preprocessing fails.
    """
    plist_file = plist.PlistFile()

    try:
      plist_file.Read(file_object)

    except OSError as exception:
      raise errors.PreProcessFail((
          f'Unable to read: {self.ARTIFACT_DEFINITION_NAME:s} with error: '
          f'{exception!s}'))

    if not plist_file.root_key:
      raise errors.PreProcessFail((
          f'Unable to read: {self.ARTIFACT_DEFINITION_NAME:s} with error: '
          f'missing root key'))

    matches = []

    self._FindKeys(plist_file.root_key, self._PLIST_KEYS, matches)
    if not matches:
      plist_keys_string = ', '.join(self._PLIST_KEYS)
      raise errors.PreProcessFail((
          f'Unable to read: {self.ARTIFACT_DEFINITION_NAME:s} with error: no '
          f'such keys: {plist_keys_string:s}.'))

    name = None
    value = None
    for name, value in matches:
      if value:
        break

    if value is None:
      plist_keys_string = ', '.join(self._PLIST_KEYS)
      raise errors.PreProcessFail((
          f'Unable to read: {self.ARTIFACT_DEFINITION_NAME:s} with error: no '
          f'values found for keys: {plist_keys_string:s}.'))

    self._ParsePlistKeyValue(mediator, name, value)

  @abc.abstractmethod
  def _ParsePlistKeyValue(self, mediator, name, value):
    """Parses a plist key value.

    Args:
      mediator (PreprocessMediator): mediates interactions between preprocess
          plugins and other components, such as storage.
      name (str): name of the plist key.
      value (str): value of the plist key.
    """


class MacOSHostnamePlugin(PlistFileArtifactPreprocessorPlugin):
  """MacOS hostname plugin."""

  ARTIFACT_DEFINITION_NAME = 'MacOSSystemConfigurationPreferencesPlistFile'

  _PLIST_KEYS = ['ComputerName', 'LocalHostName']

  def _ParsePlistKeyValue(self, mediator, name, value):
    """Parses a plist key value.

    Args:
      mediator (PreprocessMediator): mediates interactions between preprocess
          plugins and other components, such as storage.
      name (str): name of the plist key.
      value (str): value of the plist key.
    """
    if name in self._PLIST_KEYS:
      hostname_artifact = artifacts.HostnameArtifact(name=value)
      mediator.AddHostname(hostname_artifact)


class MacOSKeyboardLayoutPlugin(PlistFileArtifactPreprocessorPlugin):
  """MacOS keyboard layout plugin."""

  ARTIFACT_DEFINITION_NAME = 'MacOSKeyboardLayoutPlistFile'

  _PLIST_KEYS = ['AppleCurrentKeyboardLayoutInputSourceID']

  def _ParsePlistKeyValue(self, mediator, name, value):
    """Parses a plist key value.

    Args:
      mediator (PreprocessMediator): mediates interactions between preprocess
          plugins and other components, such as storage.
      name (str): name of the plist key.
      value (str): value of the plist key.
    """
    if name in self._PLIST_KEYS:
      if isinstance(value, (list, tuple)):
        value = value[0]

      _, _, keyboard_layout = value.rpartition('.')

      mediator.SetValue('keyboard_layout', keyboard_layout)


class MacOSSystemVersionPlugin(PlistFileArtifactPreprocessorPlugin):
  """MacOS system version information plugin."""

  ARTIFACT_DEFINITION_NAME = 'MacOSSystemVersionPlistFile'

  _PLIST_KEYS = ['ProductUserVisibleVersion']

  def _ParsePlistKeyValue(self, mediator, name, value):
    """Parses a plist key value.

    Args:
      mediator (PreprocessMediator): mediates interactions between preprocess
          plugins and other components, such as storage.
      name (str): name of the plist key.
      value (str): value of the plist key.
    """
    if name in self._PLIST_KEYS:
      mediator.SetValue('operating_system_version', value)


class MacOSTimeZonePlugin(interface.FileEntryArtifactPreprocessorPlugin):
  """MacOS time zone plugin."""

  ARTIFACT_DEFINITION_NAME = 'MacOSLocalTime'

  def _ParseFileEntry(self, mediator, file_entry):
    """Parses artifact file system data for a preprocessing attribute.

    Args:
      mediator (PreprocessMediator): mediates interactions between preprocess
          plugins and other components, such as storage.
      file_entry (dfvfs.FileEntry): file entry that contains the artifact
          value data.

    Raises:
      errors.PreProcessFail: if the preprocessing fails.
    """
    if not file_entry or not file_entry.link:
      raise errors.PreProcessFail((
          f'Unable to read: {self.ARTIFACT_DEFINITION_NAME:s} with error: not '
          f'a symbolic link'))

    _, _, time_zone = file_entry.link.partition('zoneinfo/')
    if time_zone:
      try:
        mediator.SetTimeZone(time_zone)
      except ValueError:
        mediator.ProducePreprocessingWarning(
            self.ARTIFACT_DEFINITION_NAME,
            'Unable to set time zone in knowledge base.')


class MacOSUserAccountsPlugin(interface.FileEntryArtifactPreprocessorPlugin):
  """MacOS user accounts plugin."""

  ARTIFACT_DEFINITION_NAME = 'MacOSUserPasswordHashesPlistFiles'

  _KEYS = frozenset(['gid', 'home', 'name', 'realname', 'shell', 'uid'])

  def _GetTopLevelKeys(self, top_level, keys):
    """Retrieves top-level plist keys.

    Args:
      top_level (plistlib._InternalDict): top level plist object.
      keys (set[str]): names of the top-level keys that should be retrieved.

    Returns:
      dict[str, str]: values of the requested keys or an empty dictionary if
          no corresponding top-level keys were found.
    """
    match = {}
    for key in set(keys):
      value = top_level.get(key)
      if value is not None:
        match[key] = value

    return match

  def _ParseFileEntry(self, mediator, file_entry):
    """Parses artifact file system data for a preprocessing attribute.

    Args:
      mediator (PreprocessMediator): mediates interactions between preprocess
          plugins and other components, such as storage.
      file_entry (dfvfs.FileEntry): file entry that contains the artifact
          value data.

    Raises:
      errors.PreProcessFail: if the preprocessing fails.
    """
    file_object = file_entry.GetFileObject()

    try:
      plist_file = plist.PlistFile()
      plist_file.Read(file_object)
      match = self._GetTopLevelKeys(plist_file.root_key, self._KEYS)

    except (OSError, plistlib.InvalidFileException) as exception:
      mediator.ProducePreprocessingWarning(
          self.ARTIFACT_DEFINITION_NAME,
          f'Unable to read plist with error: {exception!s}')
      return

    name = match.get('name', [None])[0]
    uid = match.get('uid', [None])[0]

    if not name or not uid:
      mediator.ProducePreprocessingWarning(
          self.ARTIFACT_DEFINITION_NAME, 'Missing name or user identifier')
      return

    user_account = artifacts.UserAccountArtifact(
        identifier=uid, username=name)
    user_account.group_identifier = match.get('gid', [None])[0]
    user_account.full_name = match.get('realname', [None])[0]
    user_account.shell = match.get('shell', [None])[0]
    user_account.user_directory = match.get('home', [None])[0]

    try:
      mediator.AddUserAccount(user_account)
    except KeyError:
      mediator.ProducePreprocessingWarning(
          self.ARTIFACT_DEFINITION_NAME,
          f'Unable to add user account: {name:s} to knowledge base.')


manager.PreprocessPluginsManager.RegisterPlugins([
    MacOSHostnamePlugin, MacOSKeyboardLayoutPlugin, MacOSSystemVersionPlugin,
    MacOSTimeZonePlugin, MacOSUserAccountsPlugin])



"""Text parser plugin for XChat log files.

Information updated 24 July 2013.

The parser applies to XChat log files. Despite their apparent
simplicity it's not straightforward to manage every possible case.
XChat tool allows users to specify how timestamp will be
encoded (using the strftime function), by letting them specify
additional separators. This parser will accept only the simplest
default English form of an XChat log file, as the following::

  **** BEGIN LOGGING AT Mon Dec 31 21:11:55 2001
  dec 31 21:11:55 --> You are now talking on #gugle
  dec 31 21:11:55 --- Topic for #gugle is plaso, nobody knows what it means
  dec 31 21:11:55 Topic for #gugle set by Kristinn
  dec 31 21:11:55 --- Joachim gives voice to fpi
  dec 31 21:11:55 *   XChat here
  dec 31 21:11:58 <fpi> ola plas-ing guys!
  dec 31 21:12:00 <Kristinn> ftw!

It could be managed the missing month/day case too, by extracting
the month/day information from the header. But the parser logic
would become intricate, since it would need to manage day transition,
chat lines crossing the midnight. From there derives the last day of
the year bug, since the parser will not manage that transition.

Moreover the strftime is locale-dependent, so month names, footer and
headers can change, even inside the same log file. Being said that, the
following will be the main logic used to parse the log files (note that
the first header *must be* '\\*\\*\\*\\* BEGIN ...' otherwise file will be
skipped).

1) Check for '\\*\\*\\*\\*'
1.1) If 'BEGIN LOGGING AT' (English)
1.1.1) Extract the YEAR
1.1.2) Generate new event start logging
1.1.3) set parsing = True
1.2) If 'END LOGGING'
1.2.1) If parsing, set parsing=False
1.2.2) If not parsing, log debug
1.2.3) Generate new event end logging
1.3) If not BEGIN|END we are facing a different language
and we don't now which language!
If parsing is True, set parsing=False and log debug
2) Not '\\*\\*\\*\\*' so we are parsing a line
2.1) If parsing = True, try to parse line and generate event
2.2) If parsing = False, skip until next good header is found

Also see:
  https://xchat.org
"""

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.lib import dateless_helper
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class XChatLogEventData(events.EventData):
  """XChat Log event data.

  Attributes:
    added_time (dfdatetime.DateTimeValues): date and time the log entry
        was added.
    nickname (str): nickname.
    text (str): text sent by nickname or other text, such as 'server' or
        'messages'.
  """

  DATA_TYPE = 'xchat:log:line'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.added_time = None
    self.nickname = None
    self.text = None


class XChatLogTextPlugin(
    interface.TextPlugin, dateless_helper.DateLessLogFormatHelper):
  """Text parser plugin for XChat log files."""

  NAME = 'xchatlog'
  DATA_FORMAT = 'XChat log file'

  ENCODING = 'utf-8'

  _ONE_OR_TWO_DIGITS = pyparsing.Word(pyparsing.nums, max=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _THREE_LETTERS = pyparsing.Word(pyparsing.alphas, exact=3)

  # TODO: Only English ASCII timestamp supported ATM, add support for others.

  _WEEKDAY = pyparsing.Group(
      pyparsing.Keyword('Sun') |
      pyparsing.Keyword('Mon') |
      pyparsing.Keyword('Tue') |
      pyparsing.Keyword('Wed') |
      pyparsing.Keyword('Thu') |
      pyparsing.Keyword('Fri') |
      pyparsing.Keyword('Sat'))

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  # Header/footer pyparsing structures.
  # Sample: "**** BEGIN LOGGING AT Mon Dec 31 21:11:55 2011".
  # Note that "BEGIN LOGGING" text is localized (default, English) and can be
  # different if XChat locale is different.

  # Header date and time values are formatted as: Mon Dec 31 21:11:55 2011
  _SECTION_HEADER_DATE_TIME = pyparsing.Group(
      _WEEKDAY + _THREE_LETTERS + _ONE_OR_TWO_DIGITS +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress(':') + _TWO_DIGITS +
      _FOUR_DIGITS)

  _LOG_ACTION = pyparsing.Group(
      pyparsing.Word(pyparsing.printables) +
      pyparsing.Word(pyparsing.printables) +
      pyparsing.Word(pyparsing.printables))

  _SECTION_HEADER_LINE = (
      pyparsing.Suppress('****') + _LOG_ACTION.set_results_name('log_action') +
      _SECTION_HEADER_DATE_TIME.set_results_name('date_time') +
      _END_OF_LINE)

  # Body (nickname, text and/or service messages) pyparsing structures.
  # Sample: "dec 31 21:11:58 <fpi> ola plas-ing guys!".

  # Date and time values are formatted as: dec 31 21:11:58
  _DATE_TIME = pyparsing.Group(
      _THREE_LETTERS + _ONE_OR_TWO_DIGITS +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress(':') + _TWO_DIGITS)

  _NICKNAME = pyparsing.QuotedString('<', endQuoteChar='>').set_results_name(
      'nickname')

  _CHAT_HISTORY_LINE = (
      _DATE_TIME.set_results_name('date_time') +
      pyparsing.Optional(_NICKNAME) +
      pyparsing.restOfLine().set_results_name('text') +
      _END_OF_LINE)

  _LINE_STRUCTURES = [
      ('chat_history_line', _CHAT_HISTORY_LINE),
      ('section_header_line', _SECTION_HEADER_LINE)]

  VERIFICATION_GRAMMAR = _SECTION_HEADER_LINE

  def __init__(self):
    """Initializes a text parser plugin."""
    super().__init__()
    self._year = None

  def _ParseLogLine(self, parser_mediator, structure):
    """Parses a log line.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      structure (pyparsing.ParseResults): structure of tokens derived from
          a line of a text file.
    """
    if not self._year:
      return

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    text = self._GetValueFromStructure(structure, 'text')
    # The text string contains multiple unnecessary whitespaces that need to
    # be removed, thus the split and re-join.
    text = ' '.join(text.split())

    event_data = XChatLogEventData()
    event_data.added_time = self._ParseTimeElements(time_elements_structure)
    event_data.nickname = self._GetValueFromStructure(structure, 'nickname')
    event_data.text = text

    parser_mediator.ProduceEventData(event_data)

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    if key == 'chat_history_line':
      self._ParseLogLine(parser_mediator, structure)

    elif key == 'section_header_line':
      self._ParseSectionHeaderLine(parser_mediator, structure)

  def _ParseSectionHeaderLine(self, parser_mediator, structure):
    """Parses a section header line.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      structure (pyparsing.ParseResults): structure of tokens derived from
          a line of a text file.
    """
    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    log_action = self._GetValueFromStructure(
        structure, 'log_action', default_value=[])

    if log_action[0] not in ('BEGIN', 'END'):
      log_action_string = ' '.join(log_action)
      parser_mediator.ProduceExtractionWarning(
          f'unsupported log action: {log_action_string:s}')
      self._year = None
      return

    event_data = XChatLogEventData()
    event_data.added_time = self._ParseTimeElements(time_elements_structure)

    if log_action[0] == 'BEGIN':
      event_data.text = 'XChat start logging'
    else:
      event_data.text = 'XChat end logging'
      self._year = None

    parser_mediator.ProduceEventData(event_data)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
          of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      if len(time_elements_structure) == 5:
        month_string, day_of_month, hours, minutes, seconds = (
            time_elements_structure)

        month = self._GetMonthFromString(month_string)

        # Use year-less helper to ensure a change in year is accounted for.
        self._UpdateYear(month)

        year = self._GetYear()

      else:
        _, month_string, day_of_month, hours, minutes, seconds, year = (
            time_elements_structure)

        month = self._GetMonthFromString(month_string)

        self._SetMonthAndYear(month, year)

      self._year = year

      time_elements_tuple = (year, month, day_of_month, hours, minutes, seconds)

      date_time = dfdatetime_time_elements.TimeElements(
          time_elements_tuple=time_elements_tuple)

      date_time.is_local_time = True

      return date_time

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    try:
      self._ParseTimeElements(time_elements_structure)
    except errors.ParseError:
      return False

    self._year = None

    return True


text_parser.TextLogParser.RegisterPlugin(XChatLogTextPlugin)
